// Copyright 1994-1998 by Jon Dart.  All Rights Reserved.
#include "moveord.h"
#include "movegen.h"
#include "scoring.h"
#include "constant.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "globals.h"
#include "config.h"

#include <iostream>

extern Move Killers1[Constants::MaxPly];
extern Move Killers2[Constants::MaxPly];

inline void swap( Move moves[], const int i, const int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i]; 
   moves[i] = tmp;
}

// move scoring routine, used only at ply 0
int Move_Ordering::score(Board &board, const Move &move)
{
   int val = 0;
   Square dest_square = DestSquare(move);
   Piece capture = board[dest_square];
   if (!IsEmptyPiece(capture))
   {
      if (!board.any_attacks(dest_square, board.OppositeSide()))
         // attack on undefended piece
      {
         val += PieceValue(capture);
      }
      else if (TypeOfMove(move) == Normal)
      {
         int est = attack_estimate(board,move);
         if (est > 0)
            val += est;
         else // losing capture
            val += 10 + (est/16);
      }
      else if (TypeOfMove(move) == Promotion)
      {
         val += PieceValue(PromoteTo(move)) - PieceValue(Pawn);
      }
      else // en-passant?
         val += 32;
   }
   Board_State stat = board.state;
   board.DoMove(move);
   if (board.any_attacks(board.KingPos(board.OppositeSide()),board.Side()))
   {
      // move into check, illegal
      val = -Constants::BIG;
   }
   else
   {
      val -= Scoring::positional_score(board,-Constants::BIG,Constants::BIG);
#ifdef _TRACE
      cout << "position=" << Scoring::positional_score(board,-Constants::BIG,Constants::BIG) << endl;
#endif
      if (board.CheckStatus() == InCheck)
         val += 25;
   }
   board.UndoMove(move,stat);
#ifdef _TRACE
   MoveImage(move,cout); cout << " score = " << val << endl;
#endif
   return val;
}

int
Move_Ordering::reorder_ply0_moves(const Board & board,
                           Move moves[], const int num_moves, 
                           const Move & best_move)
{
   if (num_moves <= 1)
      return num_moves;
   int i; int j = 0;
   for (i = 0; i < num_moves; i++)
   {
      ClearUsed(moves[i]);
      if (EqualMove(moves[i],best_move))
      {
#ifdef _TRACE
         cout << "move ";
         MoveImage(moves[i],cout);
         cout << " (pv)"<< endl;
#endif
         swap(moves,i,0); ++j;
      }
   }
   for (i=j; i< num_moves;i++) {
      if (EqualMove(moves[i],Killers1[0])) {
         swap(moves,i,j); j++;
         //break; // causes VC++98 compiler error
      }
   }
   for (i=j; i< num_moves;i++) {
      if (EqualMove(moves[i],Killers2[0])) {
         swap(moves,i,j); 
         break;
      }
   }
   return num_moves;
}

int Move_Ordering::order_ply0_moves( Board &board, Move moves[],
                                     const int num_moves, 
                                     const Move &best_move, int &easy)
{
   Scoring::init();
   int scores[Constants::MaxMoves];

   int i;
   for (i = 0; i < num_moves; i++)
   {
      if (EqualMove(moves[i],best_move))
         scores[i] = Constants::BIG;
      else
         scores[i] = score(board,moves[i]);
   }
   sort_moves(moves, scores, num_moves);
   int real_num_moves = num_moves;

   easy = 1;
   for (i=num_moves-1; i >=0; --i)
   {
      if (scores[i] == -Constants::BIG)
      {
         --real_num_moves;
      }
      else if (i >0 && scores[0]-100 < scores[i])
	easy = 0;
   }
   return real_num_moves;
}

int
Move_Ordering::order_captures(Board & board,
                              Move moves[], const int num_moves,
                              const int alpha, const int material, int ply)
{
   int scores[Constants::MaxMoves];
   int n=0;
   int bias = QSEARCH_FORWARD_PRUNE_BIAS + material;
  
   for (int i = 0; i < num_moves; i++)
   {
      PieceType capture = Capture(moves[i]);
      Square loc = StartSquare(moves[i]);
      if (capture != EmptyPiece)
      {
         int est;
         int start_val = (int)PieceValue(board[loc]);
         if (capture == King) // king capture
         {
            // There is no point sorting the
            // rest of the moves. We return only
            // this move, which signals that
            // the previous move was illegal.
            moves[0] = moves[i];
            return 1;
         }
         else
         {
            est = (int)PieceValue(capture);
         }
         extern CACHE_ALIGN Bitmap passedW[64];
         extern CACHE_ALIGN Bitmap passedB[64];
         if (start_val == PAWN_VALUE) {
            if (PromoteTo(moves[i]) != InvalidPiece)
               est += (int)PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
            else if (board.Side() == White && Bitmap::And(passedW[DestSquare(moves[i])],board.pawn_bits[Black]).is_clear())
               est += PAWN_VALUE/2;
            else if (board.Side() == Black && Bitmap::And(passedB[DestSquare(moves[i])],board.pawn_bits[White]).is_clear())
               est += PAWN_VALUE/2;
         }
         // Futility check: if capturing the piece "for free" will not
         // get our score up to alpha, discard the move
         if (bias + est >= alpha)
         {
            est -= start_val;
            if (est > 10) 
            {
               // Always consider capturing a higher-valued piece
               scores[i] = est + (int)capture;
               if (i>n){ 
                  scores[n] = scores[i];
                  moves[n] = moves[i];
               }
               ++n;
            }
            else
            {
               // This capture appears to lose, but we call attack_estimate
               // to make sure
	       est = attack_estimate(board,moves[i]);
#ifdef _TRACE
               indent(ply); cout << "move "; MoveImage(moves[i],cout);
               cout <<  " est=" << est << endl;
#endif
               if (bias + est >= alpha) {
                  scores[i] = est + (int)capture;
                  if (i>n){ 
                    scores[n] = scores[i];
                    moves[n] = moves[i];
                  }
                  ++n;
               }
            }
         }
      }
      else if (TypeOfMove(moves[i]) == Promotion)
      {
         scores[i] = PieceValue(PromoteTo(moves[i])) - PAWN_VALUE;
         if (scores[i] + PAWN_VALUE + bias  >= alpha) {
	   //            scores[i] = attack_estimate(board,moves[i]);
	   //            if (scores[i] + bias >= alpha) {
               if (i>n){ 
                  scores[n] = scores[i];
                  moves[n] = moves[i];
               }
               ++n;
	       //}
         }
      }
   }
   if (n>1){ 
      // this is generally faster than calling sort_moves, because
      // n is small.
      int swapped = 1;
      while (swapped) {
        swapped = 0;
        for (int i = 0; i < n-1; i++) {
           if (scores[i]<scores[i+1]) {
              ++swapped;
              swap(moves,i,i+1);
              int tmp = scores[i];
              scores[i] = scores[i+1];
              scores[i+1] = tmp;
           }
        }
      }
   }
   return n;
}

