// Copyright 1987-2006 by Jon Dart.  All Rights Reserved.

#include "bearing.h"
#include "search.h"
#include "scoring.h"
#include "constant.h"
#include "movegen.h"
#include "moveord.h"
#include "util.h"
#include "pinfo.h"
#include "attacke.h"
#include "log.h"
#include "globals.h"
#include "notation.h"
#include "debug.h"
#include "config.h"
#include "tbprobe.h"
//#ifdef _DEBUG
#include "legal.h"
//#endif
#ifdef _WIN32
static const UINT SEARCH_TIMER = 2;
#include <process.h>
#include <wincon.h>
#endif
#include <iostream>
#include <iomanip>
#include <limits.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

void add_pending(char *buf);

static const int Illegal = -Constants::BIG - 1000;
static int Time_Check_Interval = 5000;

static uint64 num_nodes, num_qnodes, num_moves;
static uint64 hash_hits, hash_inserts, hash_searches,
hash_replaces, hash_inserts_failed;

#if defined(EXTENDED_FUTILITY_PRUNE)
const int FUTILITY_RANGE = EXTENDED_FUTILITY_DEPTH;
#else
const int FUTILITY_RANGE = FUTILITY_DEPTH;
#endif

static const int limits[64] =
{
    0, 2,   4,  6,  8, 10, 12, 14, 16, 18,
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
    30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
    40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
    50, 51, 52, 53, 54, 55, 56, 57, 58, 59,
    60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
    70, 71, 72, 73
};

SearchContext::SearchContext():
terminate(0)
{
   for (int i = 0; i < Constants::MaxPly; i++) {
      best[i] = last_move[i] = NullMove;
      extensions[i] = 0;
   }
}

SearchContext::~SearchContext() {
}

void SearchController::init() {
    // Allocate the hash table.  Size is in kilobytes.  Guesstimate
    // a conversion to the number of hash table entries.
    int hash_entries;
    use_hash_table = options.search.hash_table_size > 0;
    hash_entries = (1000*options.search.hash_table_size)/sizeof(Hash_Entry);
    if (hash_table_size != options.search.hash_table_size) {
        // hash table was resized (GUI can do this)
        reinit();
    }
    if (use_hash_table) {
        initHash(hash_entries);
        hash_table_size = options.search.hash_table_size;
    }
}

void SearchController::cleanup(int clearHash)
{
  //    if (clearHash) {
  //        freeHash();
  //  }
}

long SearchController::calc_hash_size()
{
    // Size the hash table according to how much memory we have.
    static int sizes[12]= {
        997, 1997, 3001, 4003, 4999, 6007, 6997, 8003,
        10007, 12007, 15013, 18103
    };

    int i = 0;
#ifdef _WIN32
    SIZE_T memSize = GlobalCompact(0);
    for (i = 0; i < 11; i++)
        if (sizeof(Position_Info)*((long)sizes[i]) > memSize/16)
            break;
    return 32L*sizes[i];
#else
    return 32L*sizes[i];
#endif
}

SearchController::SearchController()
:rootSearch(NULL),
command_function(NULL),post_function(NULL),terminate_function(NULL),
pending_function(NULL),
age(0),
ratingDiff(0),
ratingFactor(0)
#ifdef SMP
,pool(ncpus-1)
#endif
{
	hash_table_size = options.search.hash_table_size;
    init();
#ifdef _WIN32
    HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
    DWORD dw;
    pipe = !GetConsoleMode(inh, &dw);
    if (!pipe) {
        SetConsoleMode(inh, dw & ~(ENABLE_LINE_INPUT |
            ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
    }
#endif
}


SearchController::~SearchController()
{
    cleanup(1);
}

Move SearchController::find_best_move(
const Board &board,
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
int background,
int isUCI,
Statistics &stat_buf,
Talk_Level t)
{
    explicit_excludes = 0;
    Move excludes[1];
    return find_best_move(board,srcType,timelimit,xtratime,plylimit,background,
        isUCI, stat_buf, t, excludes, 0);
}

Move SearchController::find_best_move(
const Board &board,
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
int background,
int isUCI,
Statistics &stat_buf,
Talk_Level t,
Move *exclude, int num_exclude)
{
    uci = isUCI;
    if (num_exclude) explicit_excludes = 1;
    // Positions are stored in the hashtable with an "age" to identify
    // which search they came from.  "Newer" positions can replace
    // "older" ones. Update the age here since we are starting a
    // new search.
    age = (age + 1) % 3;
    rootSearch = new RootSearch(board,this);
    Move best = rootSearch->ply0_search(board,srcType,timelimit,xtratime,plylimit,
                 background,stat_buf,t,exclude,num_exclude);
    delete rootSearch;
    rootSearch = NULL;
    return best;
}

void SearchController::setRatingDiff(int rdiff)
{
    ratingDiff = rdiff;
    ratingFactor = 0;
    if (ratingDiff < -100) {
        ratingFactor = Util::Max(PAWN_VALUE*((ratingDiff+100)/100),-PAWN_VALUE);
    }
    else if (ratingDiff > 100) {
        ratingFactor = Util::Min(PAWN_VALUE*((ratingDiff-100)/100),PAWN_VALUE);
    }
}


int SearchController::draw_score(const Board & board, ColorType computerSide,
       int root_total_mat) {
    int score = 0;

    // A draw is worse if there is a lot of material on the board
    if (root_total_mat > 6000) {
        score -= 16;
    }
    else if (root_total_mat > 4800) {
        score -= 16*(root_total_mat-4800)/1200;
    }
    if (board.Side() != computerSide) {
        score = -score;
    }
    int mat_diff = Scoring::material_score(board);
    // Penalize the side to move for allowing a draw if it is ahead in
    // material
    const Material &my_mat = board.getMaterial(board.Side());
    if (Util::Abs(mat_diff) >= 128 &&
    (!my_mat.no_pawns() || my_mat.value() > 2320)) {
        if (mat_diff > 0)
            score -= (mat_diff-128)/8;
        else
            score -= (mat_diff+128)/8;
    }
    // if we know the opponent's rating (which will be the case if playing
    // on ICC), factor that into the draw score - a draw against a high-rated
    // opponent is good; a draw against a lower-rated one is bad.
    if (ratingDiff != 0) {
        if (board.Side() == computerSide)
            score += ratingFactor;
        else
            score -= ratingFactor;
    }
    return score;
}

void SearchController::set_background(int b)
{
    rootSearch->set_background(b);
}

void SearchController::reinit()
{
    // After the search options change, we may need to reallocate
    // the hash table.  Free it here so that next time through init()
    // the reallocation will be done.
    freeHash();
}

#ifdef _WIN32
static int is_move(const Board &b,char *buf)
{
    if (isalpha(*buf) && isdigit(buf[1]) &&
        isalpha(buf[2]) && isdigit(buf[3]))
        return 1;
    else if (strncmp(buf,"o-o",3)==0 || strncmp(buf,"O-O",3)==0 ||
        strncmp(buf,"o-o-o",5)==0 || strncmp(buf,"O-O-O",5) ==0)
        return 1;
    else
        return Notation::Value(b,b.Side(),buf) != NullMove;
}
#endif

int SearchController::check_input(const Board &board)
{
#ifdef _WIN32
    DWORD nchar;
    DWORD dw;
    if (pipe) {
        // Use the GnuChess trick of calling PeekNamedPipe to check
        // for user input.  We need to do this because Winboard does
        // not send ^C to interrupt pondering.
        if (PeekNamedPipe(GetStdHandle(STD_INPUT_HANDLE), NULL, 0,
            NULL, &nchar, NULL)) {
            if (nchar && talkLevel == Trace) cout << "(debug) got console input nchar=" << nchar << endl;
            int nextc;
            BOOL terminate;
            for (unsigned i = 0; i < nchar; i++) {
                nextc=getc(stdin);
//                if (talkLevel == Trace) cout << "(debug) got: " << (char)nextc << endl;
                if (nextc == '\n') {
                    cmd_buf[cmd_buf_count]  = '\0';
                    cmd_buf_count = 0;
                    // call a function (supplied by the calling code) that
                    // sets terminate = 1 if we should terminate.
                    command_function(cmd_buf,terminate);
                    if (terminate) {
                        if (talkLevel == Trace) cout << "(debug) terminating" << endl;
                        return 1;
                    }
                }
                else
                    cmd_buf[cmd_buf_count++] = nextc;
            }
            // Hack alert. Sometimes we get a move with no trailing
            // carriage return.
            if (cmd_buf_count > 0) {
                cmd_buf[cmd_buf_count] = '\0';
                if (is_move(board,cmd_buf) || (strncmp(cmd_buf,"usermove",8) == 0 && is_move(board,cmd_buf+9))) {
                    if (talkLevel == Trace) cout << "(debug) got move w/o <CR>: " << cmd_buf << endl;
                    cmd_buf_count = 0;
                    command_function(cmd_buf,terminate);
                    if (terminate) {
                        return 1;
                    }
                }
            }
        }
    }
    else {
        // We are reading from the console, not connected to Winboard.
        // Not something we usually do, except for testing.
        int retVal = WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), 0);
        if (retVal == WAIT_OBJECT_0) {
            INPUT_RECORD recBuf[128];
            ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE),recBuf,
                128,&dw);
            for (unsigned i = 0; i < dw; i++) {
                if (!recBuf[i].Event.KeyEvent.bKeyDown) continue;
                char nextc=recBuf[i].Event.KeyEvent.uChar.AsciiChar;
                if (nextc == (char)13 /* carriage return */) {
                   putchar('\n');
                   cmd_buf[cmd_buf_count]  = '\0';
                   cmd_buf_count = 0;
                   BOOL terminate;
                   // call a function (supplied by the calling code) that
                   // sets terminate = 1 if we should terminate.
                   command_function(cmd_buf,terminate);
                   if (terminate) {
                       return 1;
                   }
                   break;
               }
               else {
                   putchar(nextc);
                   cmd_buf[cmd_buf_count++] = nextc;
               }
           }
       }
   }
   return 0;
#else
   // UNIX version, uses select like Crafty does
   fd_set readfds;
   struct timeval tv;
   int data;

   FD_ZERO(&readfds);
   FD_SET(fileno(stdin), &readfds);
   tv.tv_sec=0;
   tv.tv_usec=0;
   select(16, &readfds, 0, 0, &tv);
   data=FD_ISSET(fileno(stdin), &readfds);
   if (data == -1)
       cerr << "warning: select returned error. errno=" << errno << endl;
   else if (data) {
       // we have something to read
       //     cout << "(debug) got input, data = " << data << endl;
       *cmd_buf = '\0'; cmd_buf_count = 0;
       char buf[256];
       memset(buf,'\0',256);
       int bytes = read(fileno(stdin),buf,255);
       if (bytes <= 0)
           return 0;
       char *p = buf;
       int stop = 0;
       while (*p) {
           char *q = cmd_buf;
           while (*p != '\n' && *p != '\0' && cmd_buf_count < 255) {
               *q++ = *p++;
               ++cmd_buf_count;
           }
           *q = '\0';
           p++;                                      // skip over null or CR
           //cout << "(debug) cmd_buf = " << cmd_buf << endl;
           // call a function (supplied by the calling code) that
           // sets terminate = 1 if we should terminate.
           if (rootSearch->was_terminated()) {
               // already terminated the search, add this cmd to
               // the stack to be executed after the search is
               // really done
               if (pending_function) pending_function(cmd_buf);
           }
           else {
               command_function(cmd_buf,stop);
               if (stop) {
                   rootSearch->terminate_now();
               }
           }
           cmd_buf_count = 0;
       }
       return rootSearch->was_terminated();
   }
   return(0);
#endif
}

RootSearch::RootSearch(const Board &board, SearchController *c)
: Search(board,NULL,c) {
   time_up = terminated = 0;
   Move_Generator::clear_killer();
   Move_Generator::clear_history();
   LockInit(pv_lock);
}

RootSearch::~RootSearch() {
    LockDestroy(pv_lock);
}

Move RootSearch::ply0_search(const Board &board, 
Search_Type srcType,
int32 timelimit,
int32 xtratime,
int plylimit,
const int background,
Statistics &stat_buf,
Talk_Level t,
Move *exclude, int num_exclude) {

    controller->set_talk_level(t);
    time_target = timelimit;
    xtra_time = xtratime;
    ply_limit = plylimit;
    if (ply_limit > 0)
        ply_limit = Util::Min(ply_limit-1,Constants::MaxPly-1);
    type_of_search = srcType;
    num_excludes = Util::Min(MAX_PV,num_exclude);
    if (num_excludes)
       memcpy(excludes,exclude,sizeof(Move)*num_excludes);
    state = NormalState;
    stats = &stat_buf;
    stats->clear();
#ifdef EVAL_STATS
    Scoring::clearStats();
#endif
    null_depth = options.search.null_depth_reduction*DEPTH_INCREMENT;

    bkgrnd_search = background;
    Board board_copy(board);
    initial_board = board_copy;
    init(board_copy);

    Move best = NullMove;
    if (Scoring::is_legal_draw(board)) {
        // If it's a legal draw situation before we even move, then
        // just return a draw score and don't search.
        stats->state = Draw;
        stats->value = draw_score(board);
    }
    else {
        // Generate the ply 0 moves here:
        Move_Generator mg(board,0,NullMove,1);
        ply0move_count = mg.GenerateAllMoves(ply0moves,0);
        num_moves += ply0move_count;
        // Order the moves.  The ordering process removes any illegal moves
        // at ply 0, so fix the move count if necessary.
        stats->mvtot = ply0move_count =
            Move_Ordering::order_ply0_moves(initial_board, ply0moves, ply0move_count,
            NullMove,
            easy);
        // See if there is an "easy" move (a recapture that appears to
        // gain more material than any other alternative).  If so we
        // may be able to terminate the search early.
        easy = (ply0move_count > 0 &&
            board[DestSquare(ply0moves[0])] != EmptyPiece &&
            game_moves->num_moves() > 0 &&
            (DestSquare(game_moves->move(game_moves->num_moves()-1)) ==
            DestSquare(ply0moves[0])) &&
            (Capture(game_moves->move(game_moves->num_moves()-1)) != EmptyPiece &&
            Capture(ply0moves[0]) != EmptyPiece)
            );
        easy_move = ply0moves[0];
        // Do the actual search
        iterate();
        best = root->pv[0][0];
    }
    static const int end_of_game[] = {0, 0, 0, 1, 1, 1, 1};
    stats->end_of_game = end_of_game[(int)stats->state];
    if (!controller->uci && !stats->end_of_game && options.search.can_resign) {
        if (stats->depth >= 2) {
            // Use the display value - i.e. do not use score if
            // we are failing low
            if (stats->display_value <= 8-(int)Constants::BIG) {
                // TBD we can sometimes get -10000 as a score!
                // this is bogus, we should not resign.
                if (stats->display_value != -(int)Constants::BIG)
                    stats->state = Resigns;
            }
            else if (stats->display_value > 200-Constants::BIG &&
                ((100*stats->display_value)/PAWN_VALUE <= options.search.resign_threshold))
                stats->state = Resigns;
            stats->end_of_game = end_of_game[(int)stats->state];
        }
    }

    if (controller->talkLevel == Debug) {
        cout.setf(ios::fixed);
        cout << setprecision(2);
        if (stats->elapsed_time > 0) {
            char buf[128];
            stats->print_nps(buf);
            cout << buf << " nodes/second." << endl;
        }
        cout << (num_nodes-num_qnodes) << " regular nodes, " <<
            num_qnodes << " quiescence nodes." << endl;
        cout << hash_searches << " searches of hash table, " <<
            hash_hits << " successful";
        if (hash_searches != 0)
            cout << " (" <<
                (int)((100.0*(float)hash_hits)/((float)hash_searches)) <<
                " percent).";
        cout << endl;
        cout << hash_inserts << " hash entries inserted, " <<
            hash_replaces << " entries replaced, " <<
            hash_inserts_failed << " inserts failed." << endl;
#ifdef MOVE_ORDER_STATS
        cout << "move ordering: ";
        static const char *labels[] = {"1st","2nd","3rd","4th"};
        for (int i = 0; i < 4; i++) {
	  cout << setprecision(2) << labels[i] << " " <<
                (100.0*stats->move_order[i])/(float)stats->move_order_count << "% " ;
        }
        cout << endl;
#endif
        cout <<  "null cutoffs: " << stats->num_null_cuts << " (" <<
            (int)((100.0*(float)stats->num_null_cuts)/((float)stats->num_nodes)) <<
            "%)" << endl;
        cout << "pruning: " <<
#ifdef FUTILITY_PRUNE
            stats->futility_pruning << " moves (futility)" <<
#endif
#ifdef RAZORING
#ifdef FUTILITY_PRUNE
            ", " <<
#endif
            stats->num_razored << " nodes (razored)" <<
#endif
            endl;
        cout << "extensions: " <<
            stats->check_extensions << " check, " <<
            stats->recap_extensions << " recapture, " <<
            stats->pawn_extensions << " pawn, " << endl;
        cout << "            " <<
            stats->threat_extensions << " threat, " <<
            stats->forced_extensions << " forced move, " <<
            stats->rep_extensions << " repetition." << endl;
        cout << stats->tb_probes << " tablebase probes, " <<
            stats->tb_hits << " tablebase hits" << endl;
#if defined(SMP) && defined(SMP_STATS)
        cout << stats->splits << " splits," << 
	  " average thread usage=" << 1.0 + (float)(stats->threads)/(float)stats->samples << endl;
#endif
#ifdef EVAL_STATS
        extern long position_hash_probes, position_hash_hits,
            pawn_hash_probes, pawn_hash_hits,
            king_pawn_hash_probes, king_pawn_hash_hits,
            king_cover_hash_probes, king_cover_hash_hits,
            lazy_cut_chances, lazy_cuts;
        cout << "position hash hit rate = " << 100.0*(float)position_hash_hits/(float)position_hash_probes << "%" << endl;
        cout << "pawn hash hit rate     = " << 100.0*(float)pawn_hash_hits/(float)pawn_hash_probes << "%" << endl;
        cout << "king/pawn hash hit rate= " << 100.0*(float)king_pawn_hash_hits/(float)king_pawn_hash_probes << "%" << endl;
        cout << "king cover hash hit rate= " << 100.0*(float)king_cover_hash_hits/(float)king_cover_hash_probes << "%" << endl;
        if (lazy_cuts > 0L)
            cout <<"lazy cutoff in eval    = " << 100.0*(float)lazy_cuts/(float)lazy_cut_chances << "%" << endl;
#endif
        fflush(stdout);
    }
    return best;
}

void RootSearch::set_background(int b) {
    bkgrnd_search = b;
    if (!b) {
        // changing from background to foreground search,
        // update display
        show_status(initial_board,getBest());
    }
}

void RootSearch::show_status(const Board &board,const Move best)
{
    if (was_terminated())
        return;
    int ply = stats->depth;
    static char move_text[20];
    Notation::Image(board, best, move_text);
    char val[20];
    Scoring::print_score(stats->display_value,val);
    if (controller->talkLevel == Debug) {
        cout << "ply " << ply;
        if (ply < 10) cout << ".  ";
        else cout << ". ";
        cout << move_text << '\t' << stats->elapsed_time/100 <<
            " seconds.\tscore: " << val  << "\t" <<
            stats->num_nodes << " nodes." << endl;
    }
    // Post during ponder if UCI
    if (!is_background_search() || controller->uci) {
      if (controller->post_function) {
            controller->post_function(*stats);
      }
    }
}


void RootSearch::init(const Board &board) {
    if (board.state.moveCount>=90) {
        clearHash();
    }
    num_nodes = num_qnodes = num_moves = hash_hits =
        hash_searches = hash_inserts = hash_replaces = hash_inserts_failed = 0L;
    time_added = 0;
    root_total_mat = board.getMaterial(White).value() + board.getMaterial(Black).value();
    start_time = getTime();
   for (int i = 0; i < Constants::MaxPly; i++) {
      pv[i][0] = NullMove;
      pv_length[i] = 0;
   }

}


int RootSearch::check_time(const Board &board,int ply, int rank,
int iteration_depth)
{
    if (was_terminated()) {
        return 1;
    }
    if (controller->command_function && controller->check_input(board)) {
        terminate_now();
        if (controller->talkLevel == Trace)
            cout << "(debug) terminating, command input received" << endl;
        return 1;
    }
#if defined(SMP) && defined(SMP_STATS)
    stats->samples++; 
    stats->threads += controller->pool.activeCount();
#endif
    if (type_of_search != Fixed_Depth) {
        time_t current_time = (time_t)(getTime());
        // update at least the part of the stats structure that
        // is displayed in analysis mode.
        stats->elapsed_time = current_time - start_time;
        stats->num_nodes = num_nodes;
        if (current_time - (int)get_start_time() > (int)get_time_limit()) {
            if (controller->talkLevel == Trace) {
                terminate_now();
                cout << "(debug) terminating, time up" << endl;
            }
            return 1;
        }
        if (type_of_search == Fixed_Time) {
            return 0;
        }
        unsigned n = theLog->current();

        if (xtra_time > 0 &&
            type_of_search == Time_Limit &&
            (stats->elapsed_time > time_target/2) &&
        time_added == 0) {
            // see if our score has dropped a lot since our
            // last search
            if (n >= 2 && !(*theLog)[n-2].get_book_info().is_valid() &&
                (stats->value < 5*PAWN_VALUE) &&
                (stats->value > -PAWN_VALUE*5) &&
            (stats->value < (*theLog)[n-2].score() - PAWN_VALUE/3)) {
                // search more time because our score is dropping
                time_added++;
                if (controller->talkLevel == Trace) {
                    cout << "(debug) adding time" << endl;
                }
            }
        }
        // Allow terminating early if this is a recapture, appeared
        // better than everything else at initial sort time, and
        // still looks good.
        if (!time_up && easy && iteration_depth >= 5) {
            // do not do this if this is the first move out of book
            if (n>=2 && !(*theLog)[n-2].get_book_info().is_valid() &&
                time_target >= 300 &&
                // Don't do this if our score is dropping ..
                (stats->value > (*theLog)[n-2].score() - PAWN_VALUE/4) &&
                // require that we have consumed at least 1/3 of the time            // limit:
                (current_time - get_start_time())*3 > time_target) {
                if (getBest() == easy_move &&  PieceMoved(getBest()) != King) {
                    time_up = 1;
                    if (controller->talkLevel == Trace) {
                        cout << "(debug) terminating, easy move" << endl;
                    }
                    terminate_now();
                    return 1;
                }
            }
        }
    }
    if ( time_up) {
        terminate_now();
        if (controller->talkLevel == Trace)
            cout << "(debug) terminating, time up" << endl;
    }
    return time_up;
}


Move Search::getBest() const {
    return root->pv[0][0];
}

int Search::draw_score(const Board &board) {
    return controller->draw_score(board,root->side_to_move,root->root_total_mat);
}

void RootSearch::update_stats(int alpha, int beta, int multipv_count)
{
    time_t end_time = getTime();
    stats->elapsed_time = end_time - get_start_time();
    stats->num_nodes = num_nodes;
    stats->num_qnodes = num_qnodes;
    stats->num_moves = num_moves;
    stats->multipv_count = multipv_count;
    // Note: the search routine counts depth internally starting
    // at 0, but we report it starting at 1 (as most other programs do).
    stats->depth = iteration_depth+1;
    stats->value = display_score;
    // Leave the previous value intact if we have failed low.
    if (!(stats->value == alpha && iteration_depth > 0)) {
        stats->display_value = stats->value;
    }
    stats->state = state;
    *(stats->best_line_image) = '\0';
    stats->best_line[0] = NullMove;
    Board board_copy(root->initial_board);

    if (IsNull(root->pv[0][0])) {
#ifdef _TRACE
        cout << "(debug) warning: pv is null\n";
#endif
        return;
    }
    else if (root->pv_length[0] == 0) {
        return;
    }
    int i = 0;
    char *p = stats->best_line_image;
#ifdef _TRACE
	cout << "(debug) best line:" << endl;
#endif
    while (i < root->pv_length[0] && !IsNull(root->pv[0][i])) {
        ASSERT(i<Constants::MaxPly);
        stats->best_line[i] = root->pv[0][i];
        char move_image[20];
        *move_image = '\0';
#ifdef _TRACE
	MoveImage(pv[0][i],cout); cout << ' ';
#endif
        ASSERT(legal_move(board_copy,stats->best_line[i]));
        if (controller->uci) {
	   // UCI format for moves
 	   Notation::UCIMoveImage(stats->best_line[i],move_image);
	}
        else {
	  // use SAN
           Notation::Image(board_copy,stats->best_line[i],move_image);
	}
        int len = (int)strlen(move_image);
        if (p+len+1-stats->best_line_image > 255)
            break;
        strcpy(p,move_image);
        p+= len;
        strcpy(p," ");
        p++;
        board_copy.DoMove(stats->best_line[i]);
        ++i;
        int rep_count;
        if (is_draw(board_copy,rep_count,0)) {
            break;
        }
        if (root->pv_length[0] < 2 && controller->use_hash_table) {
            // get the next move from the hash table, if possible
            // (for pondering)
            Position_Info entry;
            Lock(hash_lock);
            Position_Info::ValueType result =
                searchHash(board_copy,board_copy.HashCode(rep_count),
                0,0,0,
                iteration_depth,entry);
            Unlock(hash_lock);
            if (result != Position_Info::NoHit) {
                Move hashMove = entry.best_move(board_copy);
                if (!IsNull(hashMove)) {
                    stats->best_line[i] = hashMove;
                    *p = '\0';
                    if (controller->uci) {
              	       // UCI format for moves
  	               Notation::UCIMoveImage(hashMove,move_image);
 	            }
                    else {
                       // use SAN
                       Notation::Image(board_copy,hashMove,move_image);
		    }
                    int len = (int)strlen(move_image);
                    if (p+len+1-stats->best_line_image > 255)
                        break;
                    strcpy(p,move_image);
                    p+=len;
                    ++i;
                }
                break;
            }
        }
    }
#ifdef _TRACE
    cout << endl;
#endif
    *p = '\0';

    stats->best_line[i] = NullMove;
#ifdef _TRACE
    cout << "(debug) best line image: " << stats->best_line_image << endl;
#endif
}


uint64 SearchController::get_numnodes() const
{
    // return # of nodes visited
    return rootSearch ? rootSearch->stats->num_nodes : 0;
}


void RootSearch::terminate_now()
{
    time_up = terminated = 1;
}


int RootSearch::was_terminated() const
{
    return terminated;
}

void RootSearch::iterate()
{
    // Searching may alter the board, so operate on a copy:
    Board board = initial_board;
    Time_Check_Interval = 5000;

    if (is_background_search())
        computerSide = board.OppositeSide();
    else
        computerSide = board.Side();

    context.terminate = 0;
    time_check_counter = 0;
    ply0first = NullMove;
    int value = Scoring::INVALID_SCORE;
    int tb_hit = 0;
    if (options.search.use_tablebases) {
        const Material &wMat = board.getMaterial(White);
        const Material &bMat = board.getMaterial(Black);
        int tb_score;
        if(wMat.piece_count() + bMat.piece_count() <= EGTBMenCount) {
            stats->tb_probes++;
            tb_hit = probe_tb(board, tb_score, 0);
            if (tb_hit) value = tb_score;
        }
    }
    if (value == Scoring::INVALID_SCORE) value = display_score = Scoring::evalu8(board);

    // Incrementally search the board to greater depths, up to the
    // ply limit
    for (iteration_depth = 0;
        iteration_depth <= ply_limit && !context.terminate && !time_up;
        iteration_depth++) {
        if (!controller->explicit_excludes) num_excludes = 0;
        for (int multipv_count=0; multipv_count < options.search.multipv; multipv_count++) {
           int lo_window, hi_window;
           display_score = value;
           lo_window = Util::Max(-Constants::BIG,value - INITIAL_SEARCH_WINDOW / 2);
           hi_window = Util::Min(value + INITIAL_SEARCH_WINDOW / 2,
               Constants::BIG);
#ifdef _TRACE
           cout << "iteration " << iteration_depth << " window = [" <<
               lo_window << "," << hi_window << "]" << endl;
#endif
           int failhigh = 0;
           int faillow = 0;
           int oldvalue = -Constants::BIG;
           Move oldbest = NullMove;
           for (int count = 0; 1; ++count,oldvalue = value,oldbest = getBest()) {
               if (!failhigh && !faillow)
                   ply0first = context.best[0];
               value = ply0_search(board, lo_window, hi_window,
                   DEPTH_INCREMENT*iteration_depth, 
                   excludes,num_excludes,multipv_count);
               update_stats(lo_window,hi_window,multipv_count);
               if (iteration_depth == 0 && Scoring::is_legal_draw(board)) {
                   if (controller->talkLevel == Trace)
                        cout << "draw, terminating" << endl;
                   terminate_now();
               }
               // check for forced move, but only at depth 2
               // (so we get a ponder move if possible)
               else if (iteration_depth == 2 && ply0move_count == 1) {
                   if (controller->talkLevel == Trace)
                        cout << "single legal move, terminating" << endl;
                   terminate_now();
               }
               if (!was_terminated() && (state == Checkmate || state == Stalemate)) {
                   if (controller->talkLevel == Trace)
                       cout << "terminating due to checkmate or statemate, state="
                           << (int)state << endl;
                   terminate_now();
                   break;
               }
               if (stats->elapsed_time > 100) {
                   long denom = bkgrnd_search ? 200L : 500L;
                   Time_Check_Interval = (int)(
                       (denom*stats->num_nodes)/(36L*stats->elapsed_time));
               }
               if (!context.terminate) {
                   context.terminate = check_time(board,0,0,iteration_depth);
                   if (controller->terminate_function &&
                       controller->terminate_function(*stats)) {
                       if (controller->talkLevel == Trace)
                           cout << "terminating due to program or user input" << endl;
                       terminate_now();
                   }
               }
               if (count == 0 || ((value != oldvalue) || !MovesEqual(getBest(),oldbest)))
                  show_status(board,getBest());
               if (context.terminate) {
                   break;
               }
               else {
                   if (value < hi_window && value > lo_window) {
                       // score is in the bounds of the search window,
                       // no need to re-search
                       break;
                   }
                   if (value >= hi_window) {
                       // shouldn't really happen, but put a check here.
                       // Score may reach +Constants::BIG sometimes, indicating
                       // a mating move (shouldn't be >Constants::BIG-1).
                       if (failhigh) break;
#ifdef _TRACE
                       cout << "ply 0 high cutoff, re-searching ..." << endl;
#endif
                       // high cutoff, must re-search
                       ply0first = context.last_move[0];     // so we research this move
                       hi_window = Constants::BIG;
                       lo_window = Util::Min(value,Constants::BIG-Constants::MaxPly);
                       failhigh = 1;
                       faillow = 0;
                   }
                   else if (value <= lo_window && !failhigh) {
#ifdef _TRACE
                       cout << "ply 0 low cutoff, re-searching ..." << endl;
                       cout << "value = " << value << endl;
#endif
                       hi_window = lo_window;
                       lo_window = -Constants::BIG;

                       failhigh = 0;
                       faillow = 1;
                   }
                   else {
                       // Use infinite window - next search score should be in
                       // bounds.
                       hi_window = Constants::BIG;
                       lo_window = -Constants::BIG;
                       failhigh = faillow = 0;
                   }
               }
           }
#ifdef _TRACE
           cout << iteration_depth+1 << " ply search result: ";
           MoveImage(context.best[0],cout);
           cout << " value = " << Scoring::print_score(value) << endl;
#endif
           if (options.search.multipv > 1) {
              excludes[num_excludes++] = context.best[0];
           }
           if (tb_hit && iteration_depth>=1 && !IsNull(context.best[0])) {
              if (controller->talkLevel == Trace)
                   cout << "(debug) terminating, tablebase hit" << endl;
#ifdef _TRACE
              cout << "terminating, tablebase hit" << endl;
#endif
              terminate_now();
              break;
           }
           if (value < (int)(iteration_depth*2) - Constants::BIG) {
               // We're either checkmated or we certainly will be, so
               // quit searching.
               if (controller->talkLevel == Trace)
                   cout << "(debug) terminating, low score" << endl;
#ifdef _TRACE
               cout << "terminating, low score" << endl;
#endif
               terminate_now();
               break;
           }
           else if (value >= (Constants::BIG - (int)(iteration_depth*2)
           - 1)) {
               if (iteration_depth>=2) {
                   if (controller->talkLevel == Trace)
                       cout << "(debug) terminating, high score" << endl;
#ifdef _TRACE
                   cout << "terminating, mate score" << endl;
#endif
                   terminate_now();
                   break;
               }
           }
        }
    }
}

int RootSearch::ply0_search(Board & board, int alpha, int beta,
            int depth, Move exclude [], int num_exclude, int multipv_count) {

    // implements alpha/beta search for the top most ply.  We use
    // the negascout algorithm.

    time_check_counter++;
    num_nodes++;

    int in_pv = 1;
    int in_check = 0;
    context.extensions[0] = 0;
    int num_try = 0;                              // # of legal moves tried
    int initial_alpha = alpha;

    // At this point we need to know if we are in check or not.
    in_check = board.CheckStatus() == InCheck;
    Board_State save_state = board.state;

    int try_score = alpha;
    //
    // re-sort the ply 0 moves.
    // Pv first, then captures, then killers. Killers at this ply (ply 0)
    // are moves that have previously exceeded alpha.
    //
    if (iteration_depth>0) Move_Ordering::reorder_ply0_moves(board,ply0moves, ply0move_count, ply0first);
    ASSERT(num_excludes<MAX_PV);
    for (int i = 0; i < num_exclude; i++) {
       for (int j = 0; j < ply0move_count; j++) {
           if (MovesEqual(ply0moves[j],exclude[i])) {
               SetUsed(ply0moves[j]);
           }
       }
    }

    //
    // Search the next ply
    //
    RootNodeInfo rootNode;
    rootNode.beta = rootNode.hibound = beta;
    rootNode.score = rootNode.alpha = alpha;
    rootNode.ply = 0; rootNode.depth = depth;
    rootNode.move_index = 0;
    rootNode.move_count = ply0move_count;
    rootNode.moves = ply0moves;
    for (;;) {
        Move move;
        Lock(rootNode.nodeLock);
        int ok = rootNode.move_index < ply0move_count && !rootNode.cutoff && !context.terminate; 
        if (ok) move  = ply0moves[rootNode.move_index++];
        Unlock(rootNode.nodeLock);
        if (!ok) break; 
        stats->mvleft = stats->mvtot-rootNode.move_index;
        if (controller->uci && stats->elapsed_time) {
           cout << "info currmove ";
           char image[10];
 	   Notation::UCIMoveImage(move,image);
 	   cout << image;
           cout << " currmovenumber " << rootNode.move_index << endl;
        }
        if (IsMate(move)) continue;               // move leads to mate, skip it
        if (IsUsed(move)) continue;               // move was searched already, skip it
#ifdef _TRACE
        cout << "window = [" << rootNode.alpha << ',' << rootNode.hibound << ']' <<
            " beta = " << rootNode.beta << endl;
        cout << "trying " << 0. << ". ";
        MoveImage(move,cout);
        cout << endl;
#endif
#ifdef SMP
        if (!in_pv && depth>=THREAD_SPLIT_START*DEPTH_INCREMENT) {
           // try to search this move in a different thread
           ThreadInfo *ti;
           if ((ti = controller->pool.checkOut(&rootNode)) != NULL) {
              // a thread is available, split off a new search,
              // copying state so it can run independently:
              Search *newSearch = new Search(board,this,&rootNode,
                  move,0);
              // start searching in the new thread:
              ti->start(newSearch);
#ifdef _THREAD_TRACE
	      cout << "split" << endl;
#endif
#if defined(SMP) && defined(SMP_STATS)
              stats->splits++;
#endif
              continue;
           }
        }
#endif
        context.last_move[0] = move;
        board.DoMove(move);
        ++num_try;
        // note: no extensions at ply 0
        if (depth-DEPTH_INCREMENT >= 0) {
            try_score = -search(board,-rootNode.hibound, -rootNode.alpha,
                1, depth-DEPTH_INCREMENT);
        }
        else {
            try_score = -quiesce(board, -rootNode.hibound, -rootNode.alpha,
                1, 0);
        }
        if (context.terminate) {
            board.UndoMove(move,save_state);
            break;
        }
        // We weeded out illegal moves so should never get one here
        ASSERT(try_score != Illegal);
#ifdef _TRACE
        cout << "0. ";
        MoveImage(move,cout);
        cout << ' ' << try_score;
        if (in_pv) cout << " (pv)";
        cout << endl;
#endif
        if (!in_pv &&
        ((try_score > rootNode.alpha) && (try_score < rootNode.beta))) {
            // We failed to get a cutoff and must re-search
#ifdef _TRACE
            cout << "window = [" << rootNode.alpha << "," << rootNode.beta
                 << "]" << endl;
            cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
            if (depth-DEPTH_INCREMENT >= 0) 
                try_score=-search(board,
                    -rootNode.beta,-rootNode.alpha,1,depth-DEPTH_INCREMENT);
            else
                try_score=-quiesce(board,-rootNode.beta,-rootNode.alpha,1,0);
#ifdef _TRACE
            cout << "0. ";
            MoveImage(move,cout);
            cout << ' ' << try_score;
            cout << endl;
#endif
        }
        board.UndoMove(move,save_state);
        if (context.terminate) {
            break;
        }
        // If this move leads to being mated, remove it from the
        // ply 0 move list - there is no use searching it again
        if (try_score <= -Constants::BIG+(int)iteration_depth) {
            for (int i = 0; i < ply0move_count; i++) {
                if (MovesEqual(ply0moves[i],move)) {
#ifdef _TRACE
                    cout << "removing move ";
                    MoveImage(move,cout);
                    cout << endl;
#endif
                    SetMate(ply0moves[i]);
                    continue;
                }
            }
        }
        Lock(rootNode.nodeLock);
        if (try_score > rootNode.alpha) {
            if (update_move(rootNode,move,try_score)) {
               Unlock(rootNode.nodeLock);
               break;
            }
        }
        in_pv = 0;
        rootNode.hibound = rootNode.alpha + 1;    // zero-width window
        Unlock(rootNode.nodeLock);
    }
#ifdef SMP
    // We may have other threads still searching.
    if (rootNode.cutoff || context.terminate) {
        // We just need to stop the other threads.
        // We don't care about their results.
#ifdef _THREAD_TRACE
        cout << "waiting, discarding results" << endl;
#endif
        stopAllThreads(&rootNode);
        waitForThreadCompletion(&rootNode);
    }
    else {
#ifdef _THREAD_TRACE
        cout << "waiting for results" << endl;
#endif
        // The other threads have possibly useful info for us.
        // Wait until they complete:
        waitForThreadCompletion(&rootNode);
    }
#endif
#ifdef _THREAD_TRACE
    cout << "cutoff: " << rootNode.cutoff << " best:";
    MoveImage(rootNode.best,cout); cout << " score:" << rootNode.score << endl;
#endif
    if (rootNode.cutoff)
        return try_score;
    if (in_pv && try_score < initial_alpha) {
        // we have failed low and must reset the window
//        context.best[0] = rootNode.best; // = move;
#ifdef _TRACE
        cout << "ply 0 alpha cutoff" << endl;
#endif
        if (controller->talkLevel == Debug) cout << "(debug) ply 0 alpha cutoff" << endl;
        display_score = try_score;
        return try_score;
    }

    if (num_try == 0) {
        // no moves were tried
        if (in_check) {
            if (ply0move_count == 0) {            // mate
                rootNode.score = -(Constants::BIG);
                state = Checkmate;
            }
            else {
                // We generated some moves, and some were legal,
                // but we skipped them all.
                rootNode.score = evalu8(board,alpha,beta,0);
            }
        }
        else {                                    // stalemate
            state = Stalemate;
#ifdef _TRACE
            cout << "stalemate!" << endl;
#endif
            rootNode.score = draw_score(board);
        }
    }
    else if (num_try > 0 && rootNode.score < beta) {
        // If there's only one legal move, don't search any more:
        if (num_try == 1 && rootNode.move_count == 1) {
            root->pv_length[0] = 1;
            root->pv[0][0] = context.best[0] = ply0moves[0];
            terminate_now();
        }
    }
    if (rootNode.alpha != initial_alpha && rootNode.score < beta) {
        if (num_try == 0)
            root->pv_length[0] = 0;
        else {
            update_pv(board,context.best[0],0);
            //show_status(initial_board,context.best[0]);
        }
    }
    ASSERT(rootNode.score >= -Constants::BIG && rootNode.score <= Constants::BIG);
    int score = rootNode.score;
    return score;
}

int RootSearch::update_move(NodeInfo &rootNode, Move move, int score) {
#ifdef _TRACE
    cout << "update_move, ply 0 : ";
    MoveImage(move,cout); cout  << " score=" << score;
    cout << endl;
#endif
    rootNode.best = move;
    display_score = rootNode.score = score;
    if (Capture(move)==EmptyPiece && TypeOfMove(move)!=Promotion)
        Move_Generator::set_killer(context.last_move[0], 0);
    rootNode.alpha = score;
    if (score >= rootNode.beta) {
#ifdef _TRACE
        cout << "ply 0 beta cutoff" << endl;
#endif
        if (controller->uci) {
          char buf[20];
	  Scoring::print_score_uci(score,buf);
	  cout << "info score " << buf << " lowerbound" << endl;
        }
        context.best[0] = rootNode.best;
        Lock(pv_lock);
        root->pv[0][0] = move;
        root->pv_length[0] = 1;
        rootNode.cutoff++;
        rootNode.score = score;
        Unlock(pv_lock);
        return 1; // signal cutoff
    }
    context.best[0] = move;
    return 0; // no cutoff
}

Search::Search(const Board &board, Search *parent, NodeInfo *n, Move m,int extend) 
{
    parentNode = n;
    moveToMake = m;
    ASSERT(parent != NULL);
    init(board,parent,parent->controller);
    memcpy(&context.best,&parent->context.best,sizeof(Move)*n->ply);
    memcpy(&context.last_move,&parent->context.last_move,sizeof(Move)*n->ply);
    memcpy(&context.extensions,&parent->context.extensions,sizeof(int)*n->ply);
    context.last_move[n->ply] = m;
    context.extend = extend;
}

#ifdef SMP
void Search::stopAllThreads(NodeInfo *n) {
    Lock(n->nodeLock);
    set<ThreadInfo *>::iterator it;
    for (it = n->workThreads.begin(); it != n->workThreads.end(); it++) {
       (*it)->work->terminate();
    }
    Unlock(n->nodeLock);
}

void Search::waitForThreadCompletion(NodeInfo *n) {
    Lock(n->nodeLock);
#ifdef _THREAD_TRACE
    int count = 0;
#endif
    while (!n->workThreads.empty()) {
#ifdef _THREAD_TRACE
      if (!count) cout << "wait" << endl;
      ++count;
#endif
       Unlock(n->nodeLock);

#ifdef _WIN32
       Sleep(5);       
#else
       usleep(1000);
#endif
#if defined(SMP) && defined(SMP_STATS)
       root->stats->samples++; 
       root->stats->threads += controller->pool.activeCount();
#endif
       Lock(n->nodeLock);
    }
    Unlock(n->nodeLock);
}
#endif

Search::Search(const Board &board,Search *p) 
 : moveToMake(NullMove),parentNode(NULL) {
 init(board,p,p->controller);
}

Search::Search(const Board &board, Search *p, SearchController *c) 
: moveToMake(NullMove),parentNode(NULL) {
    init(board,p,c);
}

void Search::init(const Board &board, Search *p, SearchController *c) {
    context.board = board;
    parent = p;
    controller = c;
    if (parent == NULL) {
       root = (RootSearch*)this;
    }
    else {
       root = c->rootSearch;
    }
    Scoring::init();
}


Search::~Search()
{
    cleanup();
}

void Search::update_pv(const Board &board,const Move &move,int ply)
{
    Lock(root->pv_lock);
    root->pv[ply][ply] = move;
    for (int j=ply+1;j<ply+1+root->pv_length[ply+1];j++)
        root->pv[ply][j] = root->pv[ply+1][j];
    root->pv_length[ply] = root->pv_length[ply+1]+1;
#ifdef _TRACE
   if (master()) {
#ifdef _DEBUG
       Board board_copy(board);
#endif
       indent(ply);
       cout << "update_pv, best line[" << ply << "]: ";
       for (int j2 = ply; j2 < root->pv_length[ply]+ply; j2++) {
           MoveImage(root->pv[ply][j2],cout); cout << " " << (flush);
#ifdef _DEBUG
           ASSERT(legal_move(board_copy,root->pv[ply][j2]));
           board_copy.DoMove(root->pv[ply][j2]);
#endif
       }
       cout << endl;
    }
#endif
    Unlock(root->pv_lock);
}

int Search::is_draw( const Board &board, int &rep_count, int ply) {
    // First check for draw by repetition
    int entries = board.state.moveCount-2;
    hash_t to_match = board.HashCode();
    rep_count = 0;
    if (entries >= 0) {
        for (hash_t *rep_list=board.rep_list_head-3;
            entries>=0;
            rep_list-=2,entries-=2) {
            //#ifdef _TRACE
            //    indent(ply);
            //       cout << " comparing " <<
            //    (hex) << Bitmap(*rep_list).hivalue() <<
            //       Bitmap(*rep_list).lovalue() << (dec) << endl;
            //#endif
            if (*rep_list == to_match) {
                rep_count++;
                if (rep_count >= 2) {// ||
//                    (rep_count == 1 && (ply<=4)/*board.rep_list_head-rep_list <= ply*/)) {
//                    (rep_count == 1 && board.rep_list_head-rep_list <= ply)) {
#ifdef _TRACE
                   if (master()) {
                      indent(ply); cout << "rep count=" << rep_count << endl;
                   }
#endif
                   return 1;
                }
            }
        }
    }
#ifdef _TRACE
    if (rep_count && master()) {
        indent(ply); cout << "rep count=" << rep_count << endl;
    }
#endif
    // check the 50 move rule
    if (board.state.moveCount >= 100) {
       return 1;
    }
    // check for insufficient material and other drawing situations
    return Scoring::material_draw(board) || Scoring::is_theoretical_draw(board);
}


int Search::evalu8(const Board & board, int alpha, int beta, int ply)
{
    // Evaluate a board position.
    int rep_count;
    if (is_draw(board,rep_count,ply))
        return draw_score(board);
    int score = Scoring::material_score(board);
    int pos_score = Scoring::positional_score(board,-Constants::BIG,
        Constants::BIG)
        + Scoring::asymm_evalu8(board,root->computerSide);
    score += pos_score;
    return score;
}

int Search::
quiesce(Board & board, int alpha, int beta,
        int ply, int depth)
{
    // recursive function, implements quiescence search.
    //
    ASSERT(ply < Constants::MaxPly);
    num_nodes++; root->time_check_counter++; num_qnodes++;
    root->pv_length[ply] = 0;
    if (ply >= Constants::MaxPly-1) {
        return evalu8(board,alpha,beta,ply);
    }
#ifdef _TRACE
    if (master()) {
      indent(ply); cout << "window [" << alpha << "," << beta << "]" << endl;
    }
#endif
    if (board.CheckStatus(context.last_move[ply-1])==InCheck) {
      if (Bearing::is_pinned(board,board.OppositeSide(),
			     MakePiece(PieceMoved(context.last_move[ply-1]),board.Side()),
			     StartSquare(context.last_move[ply-1]),
			     DestSquare(context.last_move[ply-1])))
	return -Illegal;
      else
        return qsearch_check(board,alpha,beta,ply,depth);
    }
    else
      return qsearch_no_check(board,alpha,beta,ply,depth);
}

int Search::qsearch_no_check(Board & board, int alpha, int beta,
			     int ply, int depth) {
    int stand_pat_score;
    int mat_score; 
    // Establish a default score.  This score is returned if no
    // captures are generated, or if no captures generate a better
    // score (since we generally can choose whether or not to capture).
    mat_score = Scoring::material_score(board);
    if (mat_score - 5*PAWN_VALUE > beta) {
#ifdef _TRACE
        if (master()) {
           indent(ply); cout << "** CUTOFF ** (material)" << endl;;
        }
#endif
        return mat_score;
    }
    int bitscore;
    if (board.getMaterial(White).material_level()==0 &&
        board.getMaterial(Black).material_level()==0 &&
        ((bitscore=Scoring::tryBitbase(board))!=Scoring::INVALID_SCORE)) {
         stand_pat_score = bitscore;
    }
    else {
        stand_pat_score = mat_score+Scoring::positional_score(board,alpha-mat_score,beta-mat_score)
            + Scoring::asymm_evalu8(board,root->computerSide);
    }
#ifdef _TRACE
    if (master()) {
       indent(ply);
       cout << "cap score = " << stand_pat_score << endl;
    }
#endif
    if (stand_pat_score > alpha) {
        if (stand_pat_score >= beta) {
#ifdef _TRACE
            if (master()) {
               indent(ply); cout << "**CUTOFF**" << endl;;
            }
#endif
            return stand_pat_score;
        }
        alpha = stand_pat_score;
    }

    int score = alpha;
    Move moves_generated[Constants::MaxMoves];
    Move *moves = moves_generated;
    Move_Generator mg(board, ply, NullMove, master());
    // generate all the capture moves
    int move_count = mg.GenerateCaptures(moves);
    num_moves += move_count;
#ifdef _TRACE
    int move_count1 = move_count;
#endif
    move_count = Move_Ordering::order_captures(board, moves, move_count, alpha,
          stand_pat_score,ply);
#ifdef _TRACE
    if (master()) {
        indent(ply); cout << move_count1 << " capture moves, ";
       cout << move_count << " after pruning." << endl;
    }
    if (move_count == 0) {
       return score;                             // no capture moves
    }
#endif
    Board_State save_state = board.state;

    int initial_alpha = alpha;
    int move_index = 0;
    int num_try = 0;                              // # of legal moves tried
    int try_score;
    context.best[ply] = NullMove;

    const int swag = QSEARCH_FORWARD_PRUNE_BIAS+stand_pat_score;
    while (move_index < move_count) {
        Move emove = moves[move_index++];
        if (Capture(emove) == King) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
        // We already did forward pruning in order_captures, but if alpha has
        // increased we have an opportunity to do it again.
	if (alpha != initial_alpha) {
            int est = PieceValue(Capture(emove));
            if (PromoteTo(emove) != InvalidPiece)
                est += (int)PieceValue(PromoteTo(emove)) - PAWN_VALUE;
            if (swag + est < alpha) {
                continue;
            }
            else {
                est = est - PieceValue(PieceMoved(emove));
                if (est <= 10 && Capture(emove) != EmptyPiece) {
                    est = attack_estimate(board,emove);
                    if (swag + est < alpha) {
                        continue;
                    }
                }
            }
	}
#ifdef _TRACE
        if (master()) {
          indent(ply);
          cout << "trying " << ply << ". ";
          MoveImage(emove,cout);
          cout << endl;
        }
#endif
        context.last_move[ply] = emove;
        board.DoMove(emove);
        try_score = -quiesce(board, -beta, -alpha, ply+1, depth-1);
        board.UndoMove(emove,save_state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << ply << ". ";
              MoveImage(emove,cout);
              cout << ' ' << try_score << endl;
            }
#endif
            num_try++;
            if (try_score > alpha) {
                score = try_score;
                if (try_score >= beta)
                    break;
                context.best[ply] = emove;
                alpha = try_score;
                if (score >= Constants::BIG-1-ply)
                    break;                        // mating move found
            }
        }
#ifdef _TRACE
        else if (master()) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(emove,cout);
            cout << " (illegal)" << endl;
        }
#endif
    }
    if (score >= beta) {
#ifdef _TRACE
        if (master()) {
           indent(ply);
           cout << "**CUTOFF**" << endl;
        }
#endif
        return score;
    }
    if (depth >= 1-options.search.checks_in_qsearch) {
       // Do checks in qsearch 
       Move_Generator mg(board, ply, NullMove, master());
       move_count = mg.GenerateChecks(moves);
       move_index = 0;
#ifdef _TRACE
       if (master()) {
	     if (move_count) {
             indent(ply);
		     cout << move_count << " checks generated" << endl;;
		     cout << board << endl;
	     }
       }
#endif
       while (move_index < move_count) {
           Move emove = moves[move_index++];
#ifdef _TRACE
           if (master()) {
             indent(ply);
             cout << "trying " << ply << ". ";
             MoveImage(emove,cout);
             cout << endl;
           }
#endif
           context.last_move[ply] = emove;
           board.DoMove(emove);
           try_score = -quiesce(board, -beta, -alpha, ply+1, depth-1);
           board.UndoMove(emove,save_state);
           if (try_score != Illegal) {
#ifdef _TRACE
               if (master()) {
                 indent(ply);
                 cout << ply << ". ";
                 MoveImage(emove,cout);
                 cout << ' ' << try_score << endl;
               }
#endif
               num_try++;
               if (try_score > alpha) {
                   score = try_score;
                   if (try_score >= beta)
                       break;
                   context.best[ply] = emove;
                   alpha = try_score;
                   if (score >= Constants::BIG-1-ply)
                       break;                        // mating move found
               }
           }
#ifdef _TRACE
           else if (master()) {
               indent(ply);
               cout << ply << ". ";
               MoveImage(emove,cout);
               cout << " (illegal)" << endl;
           }
#endif
       }
       if (score >= beta) {
#ifdef _TRACE
           if (master()) {
              indent(ply);
              cout << "**CUTOFF**" << endl;
           }
#endif
           return score;
       }
    }
    if (alpha != initial_alpha && score < beta) {
        if (IsNull(context.best[ply]))
            root->pv_length[ply] = 0;
        else {
            update_pv(board,context.best[ply],ply);
        }
    }
    return score;
}

int Search::qsearch_check(Board & board, int alpha, int beta,
			     int ply, int depth) {
    int score = alpha;

    int initial_alpha = alpha;
    int num_try = 0;                              // # of legal moves tried
    int try_score;
    context.best[ply] = NullMove;
    
    Board_State save_state = board.state;
    Move_Generator mg(board, ply, NullMove, master());

    Move emove;
    while ((emove = mg.NextEvasion()) != NullMove) {
        if (Capture(emove) == King) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
#ifdef _TRACE
        if (master()) {
          indent(ply);
          cout << "trying " << ply << ". ";
          MoveImage(emove,cout);
          cout << endl;
        }
#endif
        context.last_move[ply] = emove;
        board.DoMove(emove);
        try_score = -quiesce(board, -beta, -alpha, ply+1, depth-1);
        board.UndoMove(emove,save_state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
              indent(ply);
              cout << ply << ". ";
              MoveImage(emove,cout);
              cout << ' ' << try_score << endl;
            }
#endif
            num_try++;
            if (try_score > alpha) {
                score = try_score;
                if (try_score >= beta)
                    break;
                context.best[ply] = emove;
                alpha = try_score;
                if (score >= Constants::BIG-1-ply)
                    break;                        // mating move found
            }
        }
#ifdef _TRACE
        else if (master()) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(emove,cout);
            cout << " (illegal)" << endl;
        }
#endif
    }
    if (score >= beta) {
#ifdef _TRACE
        if (master()) {
           indent(ply);
           cout << "**CUTOFF**" << endl;
        }
#endif
        return score;
    }
    if (num_try == 0) {
        return -(Constants::BIG - ply); // mate
    }
    if (alpha != initial_alpha && score < beta) {
        if (IsNull(context.best[ply]))
            root->pv_length[ply] = 0;
        else {
            update_pv(board,context.best[ply],ply);
        }
    }
    return score;
}

void Search::cleanup() {
}


#ifdef _DEBUG
// To reduce the chance of a hash collision, we do some elementary
// testing on moves returned by the hash table, to help weed out
// any that are invalid in the current board position ..
static int hash_move_valid( const Board &board, const Move &move)
{
    if (IsNull(move))
        return 1;
    if (StartSquare(move) == DestSquare(move))
        return 0;
    Piece p = board[StartSquare(move)];
    if (TypeOfPiece(p) != EmptyPiece && PieceColor(p) == board.Side()) {
        p = board[DestSquare(move)];
        if (IsEmptyPiece(p) || PieceColor(p) != board.Side()) {
            if (TypeOfMove(move) == EnPassant) {
                return board.EnPassantSq() != InvalidSquare;
            }
            else if (TypeOfMove(move) == KCastle) {
                return (board.CastleStatus(board.Side()) == CanCastleKSide) ||
                    (board.CastleStatus(board.Side()) == CanCastleEitherSide);
            }
            else if (TypeOfMove(move) == QCastle) {
                return (board.CastleStatus(board.Side()) == CanCastleQSide) ||
                    (board.CastleStatus(board.Side()) == CanCastleEitherSide);
            }
            else
                return 1;
        }
        else
            return 0;
    }
    else
        return 0;
}
#endif

static int FORCEINLINE passed_pawn_push(const Board &board, Move move) {
  extern CACHE_ALIGN Bitmap passedW[64],passedB[64];
  int passed = 0;
  if (PieceMoved(move)==Pawn) {
     if (board.Side() == White) 
	passed = Bitmap::And(passedW[DestSquare(move)],board.pawn_bits[Black]).is_clear();
     else
	passed = Bitmap::And(passedB[DestSquare(move)],board.pawn_bits[White]).is_clear();
  }
  return passed;
}

static FORCEINLINE int pruneOk(const Board &board, int alpha,
int depth,int in_check, Move move, int &fmax, int &mat_score)
{
    int fmargin;
    if (depth <= FUTILITY_RANGE && !in_check && !(PieceMoved(move)==Pawn && passed_pawn_push(board,DestSquare(move)))) {
        fmargin = (depth>DEPTH_INCREMENT) ? EXTENDED_FUTILITY_MARGIN : FUTILITY_MARGIN;
        if (mat_score == -Constants::BIG)
	  mat_score = Scoring::material_score(board);
        if (fmargin + mat_score  < alpha) {
            fmax = fmargin + mat_score;
            return 1;
        }
    }
    return 0;
}

extern int hashSize;

#ifdef SMP
void Search::searchSMP(ThreadInfo *ti) {
#ifdef _THREAD_TRACE
    cout <<"searchSMP, count = " << ti->parentNode->workThreads.size() << endl;
#endif
    Move move = ti->work->moveToMake;
    while (!context.terminate) {
       Lock(ti->parentNode->nodeLock);
       int depth = ti->parentNode->depth;
       int alpha = ti->parentNode->alpha;
       int beta = ti->parentNode->beta;
       int hibound  = ti->parentNode->hibound;
       int ply = ti->parentNode->ply;
       int cutoff = ti->parentNode->cutoff;
       context.last_move[ply] = move;
       Unlock(ti->parentNode->nodeLock);
       if (cutoff) break;
       int try_score;
       Board_State state(context.board.state);
       ti->work->context.board.DoMove(move);
       // check for legality (unless at ply 0)
       if (ply > 0) {
           if (context.board.CheckStatus() != InCheck && 
               context.board.any_attacks(context.board.KingPos(
               context.board.OppositeSide()),context.board.Side())) {
               // move is illegal
               controller->pool.checkIn(ti);
               return;
           }
       }
       int extend = context.extend;
       if (depth+extend-DEPTH_INCREMENT >= 0) {
           if (ply == 0) 
              try_score = -search(context.board,-hibound, -alpha,
                  ply+1, depth+extend-DEPTH_INCREMENT);
           else
              try_score = -search(context.board,-hibound, -alpha,
                  ply+1, depth+extend-DEPTH_INCREMENT);
       }
       else {
           try_score = -quiesce(context.board, -hibound, -alpha,
               ply+1, 0);
       }
       if (context.terminate) {
           controller->pool.checkIn(ti);
           return;
       }
       if (try_score > alpha && try_score < beta) {
          if (depth+extend-DEPTH_INCREMENT >= 0) {
              if (ply == 0) 
                 try_score = -search(context.board,-beta, -alpha,
                     ply+1, depth+extend-DEPTH_INCREMENT);
              else
   
                 try_score = -search(context.board,-beta, -alpha,
                     ply+1, depth+extend-DEPTH_INCREMENT);
          }
          else {
              try_score = -quiesce(context.board, -beta, -alpha,
                  ply+1, 0);
          }
       }
       if ((context.extensions[ply] & HPRUNE) && try_score >= beta) {
          extend = 0;
#ifdef _TRACE
          indent(ply); cout << "verifying history pruning, depth " << depth-DEPTH_INCREMENT << endl;
#endif
          // verify history pruning
          if (depth+extend-DEPTH_INCREMENT >= 0) {
              try_score = -search(context.board, -beta, -alpha,
                  ply+1, depth-DEPTH_INCREMENT);
          }
          else {
              try_score = -quiesce(context.board, -beta, -alpha,
                  ply+1, 0);
          }
       }
#ifdef _THREAD_TRACE
       cout << "SMP search, ply " << ply << ": "; MoveImage(move,cout);
       cout << " score=" << try_score << " alpha=" << alpha << " beta=" <<
               beta << endl;
#endif
       Lock(ti->parentNode->nodeLock);
#ifdef _THREAD_TRACE
       cout <<"score=" << try_score << endl; 
#endif
       if (try_score > alpha) {
          // search produced a new best move or cutoff, update parent node
          if (ply == 0) {
             ti->work->parent->update_move(*parentNode,context.last_move[ply],try_score);
          }
          else
             ti->work->parent->Search::update_move(*parentNode,context.last_move[ply],try_score);
       }
       Unlock(ti->parentNode->nodeLock);
       // See if we can grab another node to search
       if (ti->parentNode->ply == 0 && !ti->parentNode->cutoff) {
          context.board.UndoMove(move,state);
          Lock(ti->parentNode->nodeLock);
          RootNodeInfo *rni = (RootNodeInfo*)ti->parentNode;
          if (rni->move_index < rni->move_count) {
             Move *loc = &(rni->moves[rni->move_index++]);
             move = *loc;
#ifdef _THREAD_TRACE
             cout << " grabbed " << rni->move_index-1 << "/" << rni->move_count << " " ; MoveImage(move,cout); cout << endl;
#endif
             Unlock(ti->parentNode->nodeLock);
          }
          else {
             Unlock(ti->parentNode->nodeLock);
             break;
          }
       }
       else
          break;
    }
    // this search is done, make this thread available for
    // other tasks:
#ifdef _THREAD_TRACE
    cout << "checkin " << endl;
#endif
    controller->pool.checkIn(ti);
#ifdef _THREAD_TRACE
    cout << "out of SearchSMP" << endl;
#endif
}
#endif

int Search::update_move(NodeInfo &node, Move move, int score) {
#ifdef _TRACE
    cout << "update_move, ply " << node.ply << ": ";
	for (int i=0;i<=node.ply;i++) {
       MoveImage(context.last_move[i],cout);
       cout << " ";
    }
    cout << endl;
#endif
    ASSERT(node.ply);
    node.score = score;
    //if (Capture(move)==EmptyPiece && TypeOfMove(move)!=Promotion)
    //    set_killer(context.last_move[0], 0);
    node.best = context.best[node.ply] = move;
    node.alpha = score;
    if (score >= node.beta) {
#ifdef _TRACE
        indent(node.ply); cout << "beta cutoff" << endl;
#endif
        node.cutoff++;
        return 1; // signal cutoff
    }
    return 0; // no cutoff
}

int Search::search(Board & board, int alpha, int beta,
                   int ply, int depth) {
    // recursive function, implements alpha/beta search.  We use
    // the PVS (principal variation search) variant of alpha/beta.
    if (master() && root->time_check_counter >= Time_Check_Interval) {
        root->time_check_counter = 0;
        context.terminate = root->check_time(board,ply,0,root->iteration_depth);
    }
    ASSERT(ply < Constants::MaxPly);
    root->time_check_counter++;
    num_nodes++;
    context.best[ply] = NullMove;                         // move we will return
    if (context.terminate) {
        return alpha;
    }
    int in_check = 0;
    int threat = 0;
    int rep_count;
#ifdef FUTILITY_PRUNE
    int est;
#endif
    if (is_draw(board,rep_count,ply)) {
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "draw!" << endl;
        }
#endif
        int score = draw_score(board);
        if (score > alpha && score < beta) {
            root->pv[ply][ply] = NullMove;
            root->pv_length[ply] = 0;
        }
        return score;
    }
    else if (ply >= Constants::MaxPly-1) {
        return evalu8(board,alpha,beta,ply);
    }
    Move hash_move = NullMove;
    Position_Info hash_entry;
    Position_Info::ValueType result = Position_Info::NoHit;
    // Note: query the hash table before the tablebases, since TB info may
    // be cached. 
    if (controller->use_hash_table) {
        // Search the hash table to see if we have hit this
        // position before.
        //
        Lock(hash_lock);
        result = searchHash(board,board.HashCode(rep_count),
            alpha,beta,ply,depth,hash_entry);
        Unlock(hash_lock);
        hash_searches++;
#ifdef _DEBUG
        if (result != Position_Info::NoHit) {
            // Found the position in the hash table.
            hash_move = hash_entry.best_move(board);
            if (!IsNull(hash_move) && !legal_move(board,hash_move)) {
                ASSERT(0);
            }
        }
#endif
    }
    int using_tb = 0;
    if (options.search.use_tablebases) {
        const Material &wMat = board.getMaterial(White);
        const Material &bMat = board.getMaterial(Black);
        using_tb = (wMat.piece_count() + bMat.piece_count() <= EGTBMenCount) &&
           ((ply <= Util::Max(1,3*root->iteration_depth/4) || 
             Capture(context.last_move[ply-1])!=EmptyPiece));
    }
    if (result != Position_Info::NoHit) {
        hash_hits++;
        // if using TBs at this ply, do not pull a non-TB entry out of
        // the hashtable. Also do not allow cut of a full node from a
        // cut hash entry.
        if ((using_tb && depth != Constants::MaxPly*DEPTH_INCREMENT) ||
            (alpha+1>beta && !hash_entry.isFull()))
   	   result = Position_Info::Invalid;
        int value = hash_entry.value();
        switch (result) {
            case Position_Info::Valid:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = E" <<
                        " alpha = " << alpha << " beta = " << beta <<
                        " value = " << value << endl;
                }
#endif
                if (value > alpha && value < beta) {
                    root->pv[ply][ply] = hash_entry.best_move(board);
#ifdef _TRACE
                    if (master()) {
                        indent(ply); cout << "best line[ply][ply] = ";
                        MoveImage(hash_entry.best_move(board),cout);
                        cout << endl;
                    }
#endif
                    root->pv_length[ply] = 1;
                }
                return value;
            case Position_Info::UpperBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = U" <<
                        " alpha = " << alpha << " beta = " << beta <<
                        " value = " << value << endl;
                }
#endif
                return value;                     // cutoff
            case Position_Info::LowerBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = L" <<
                        " alpha = " << alpha << " beta = " << beta <<
                        " value = " << value << endl;
                }
#endif
                return value;                     // cutoff
            default:
                // Note: hash move & flags may be usable even if score is not usable
                hash_move = hash_entry.best_move(board);
                board.set_check_status(hash_entry.in_check() ?
                    InCheck : NotInCheck);
                break;
        }
    }
    if (using_tb) {
        int tb_hit = 0, tb_score;
        root->stats->tb_probes++;
        tb_hit = probe_tb(board, tb_score, ply);
        if (tb_hit) {
#ifdef _TRACE
            if (master()) {
                indent(ply); cout << "EGTB hit: score " << tb_score << endl;
            }
#endif
            root->stats->tb_hits++;
            int score = tb_score;
            root->pv[ply][ply] = NullMove;
            root->pv_length[ply] = 0;
            // insert TB info in hash table.
            if (score <= -Constants::BIG+200) {
                score -= ply;
            }
            else if (score >= Constants::BIG-200) {
                score += ply;
            }
#ifdef _TRACE
            if (master() && tb_score != score) {
                indent(ply); cout << "adjusted score " << score << endl;
            }
#endif
            // Put it in with a large depth so we will not
            // overwrite - this entry is "exact" at all
            // search depths, so effectively its depth is infinite.
            Hash_Entry new_pos(board.HashCode(rep_count),
                Constants::MaxPly*DEPTH_INCREMENT,
                controller->age,
                Position_Info::Valid, 0,
                board.CheckStatus() == InCheck,
                score, 1, NullMove
                );
            Lock(hash_lock);
            switch (storeHash(new_pos)) {
                case 0: hash_inserts++; break;
                case 1: hash_replaces++; break;
                case 2: hash_inserts_failed++; break;
                default:  break;
            }
            Unlock(hash_lock);
            return tb_score; // unadjusted score
        }
    }
#ifdef _TRACE
    if (master() && !IsNull(hash_move)) {
        indent(ply);
        cout << "hash move = ";
        MoveImage(hash_move,cout);
        cout << endl;
    }
#endif
    int initial_alpha = alpha;
    // At this point we need to know if we are in check or not.
    in_check = board.CheckStatus(context.last_move[ply-1]) == InCheck;
#ifdef _DEBUG	
	if (!in_check) {
		if(board.calc_attacks(board.KingPos(board.Side()),board.OppositeSide())!=Bitmap(0)) {
			cout << board << endl;
			MoveImage(context.last_move[ply-1],cout);
			cout << endl;
			ASSERT(0);
		}
	}
        else {
		if(board.calc_attacks(board.KingPos(board.Side()),board.OppositeSide())==Bitmap(0)) {
			cout << board << endl;
			MoveImage(context.last_move[ply-1],cout);
			cout << endl;
			ASSERT(0);
		}
	}
#endif
#ifdef _TRACE
    if (master() && in_check) { indent(ply); cout << "in_check=" << in_check << endl;}
#endif
#if defined(FUTILITY_PRUNE) || defined(RAZORING)
    int mat_score=-Constants::BIG;
#endif
#if defined(FUTILITY_PRUNE)
    int fpruned_moves=0;
#endif
#ifdef RAZORING
    int razor = 0;
    if (!in_check && board.getMaterial(board.OppositeSide()).piece_count() >= 3
        &&  depth <= RAZOR_DEPTH && depth > RAZOR_DEPTH+RAZOR_EXTENSION) {
        if (mat_score == -Constants::BIG)
            mat_score = Scoring::material_score(board);
        razor = (mat_score + RAZOR_MARGIN) < alpha;
    }
#endif

#ifdef MOVE_ORDER_STATS
    int best_count = 0;
#endif
    int move_count = 0;
    int num_try = 0;                              // # of legal moves tried

    NodeInfo node;
    node.alpha = node.score = alpha; node.beta = node.hibound = beta;
    node.ply = ply; node.depth = depth;

    memcpy(&context.states[ply],&board.state,sizeof(Board_State));

    // Try to get a fast cutoff using a "null move".  Skip if the side
    // to move is in check or if material balance is low enough that
    // zugzwang is a possibility.
    int nullOk = 0;
    int n = 0;
    if (root->null_depth && !in_check && (beta==alpha+1) &&
        !IsNull(context.last_move[ply-1]) &&
        !board.getMaterial(Black).no_pieces() &&
        !board.getMaterial(White).no_pieces()) {
        n = board.getMaterial(board.Side()).has_queen() ? 1 : 2;
        nullOk =
            (board.getMaterial(board.Side()).majorPieceCount() >= n);
    }
    if (nullOk) {
        context.last_move[ply] = NullMove;
        board.DoNull();
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "trying " << ply << ". " << "(null)" << endl;
        }
#endif
        int nu_depth = depth-root->null_depth-DEPTH_INCREMENT;
#ifdef ADAPTIVE_NULL
        if (depth >= 9*DEPTH_INCREMENT ||
            (depth >= 6*DEPTH_INCREMENT &&
            board.getMaterial(board.Side()).majorPieceCount() >= 3))
            nu_depth -= DEPTH_INCREMENT;          // R=3
#endif
        int nscore;
        if (nu_depth >= 0)
            nscore = -search(board, -beta, -alpha,
                ply+1, nu_depth);
        else
            nscore = -quiesce(board, -beta, -alpha,
                ply+1, 0);
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << ply << ". " << "(null)" << ' ' << nscore << endl;
        }
#endif
        board.UndoNull(context.states[ply]);
        if (context.terminate) {
            return alpha;
        }
        else if (nscore >= beta) {                // cutoff
#ifdef VERIFY_NULL_SEARCH
	   if (board.getMaterial(board.Side()).material_level() <= 6) {
                // in endgame, verify null cutoff with reduced-depth search
                // (idea from Dieter Buerssner)
                nu_depth = depth-2*DEPTH_INCREMENT;
                if (nu_depth >= 0)
                    nscore = search(board, alpha, beta,
                        ply+1, nu_depth);
                else
                    nscore = quiesce(board, alpha, beta,
                        ply+1, 0);
            }
            if (nscore >= beta)                   // null cutoff
#endif
            {
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "**CUTOFF**" << endl;
                }
#endif
                root->stats->num_null_cuts++;
                node.score = nscore;
                goto search_end;
            }
        }
        if (nscore < 100-Constants::BIG) {
            threat = THREAT_EXTENSION;
        }
    }

    // Use "internal iterative deepening" to get an initial move to try if
    // there is no hash move .. an idea from Crafty.
    if (IsNull(hash_move) && depth > 2*DEPTH_INCREMENT) {//  && (beta != alpha+1)) {
        int d;
	d = Util::Min(depth-2*DEPTH_INCREMENT,depth/2);
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "== deepening, depth = " << d
				<< endl;
        }
#endif
        // call special search routine. Return value not significant.
        dsearch(board,alpha,beta,ply,d);
        if (context.terminate) {
            return alpha;
        }
#ifdef _TRACE
        if (master()) { 
           indent(ply); cout << "== deepening done.";
        }
#endif
        hash_move = context.best[ply];
#ifdef _DEBUG
        if (!IsNull(hash_move)) {
          ASSERT(legal_move(board,hash_move));
        }
#endif
#ifdef _TRACE
        if (master()) {
            if (!IsNull(hash_move)) {
                indent(ply); cout << "  hash_move = ";
                MoveImage(hash_move,cout);
            }
            cout << endl;
        }
#endif
    }
    {
        Move_Generator mg(board, ply, hash_move, master());
        int try_score;
        // place to taper off extensions:
        const int limit = ply > limits[root->iteration_depth];
        //
        // Now we are ready to loop through the moves from this position
        //
        int first = 1;
        while (!node.cutoff) {
            Move move = in_check ? mg.NextEvasion() : mg.NextMove();
            if (IsNull(move)) break;
            if (IsUsed(move)) continue;
            move_count++;
            num_moves++;
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "trying " << ply << ". ";
                MoveImage(move,cout);
                if (first) cout << "(pv)";
                cout << " [" << node.alpha << "," << node.beta << "]";
                cout << endl;
            }
#endif
            context.last_move[ply] = move;
            CheckStatusType in_check_after_move = board.wouldCheck(move);
            int extend = 0;
            // see if we should apply any extensions at this node.
            context.extensions[ply] = 0;
            if (options.search.repetition_extensions && rep_count==1) {
                context.extensions[ply] |= REPETITION;
                extend += REPETITION;
                root->stats->rep_extensions++;
            }
            if (options.search.forced_extensions && mg.forced()) {
                // one single reply to check
                extend += FORCED_EXTENSION;
                root->stats->forced_extensions++;
            }
            else if (options.search.check_extensions && in_check_after_move == InCheck) {
                context.extensions[ply] |= CHECK;
                extend += CHECK_EXTENSION;
                root->stats->check_extensions++;
            }
            else if (options.search.pawn_push_extensions && 
                PieceMoved(move) == Pawn &&
                Rank(DestSquare(move),board.Side())==7) {
                 context.extensions[ply] |= PAWN_PUSH;
                 root->stats->pawn_extensions++;
            }
            else if (options.search.recapture_extensions &&
                Capture(move) != EmptyPiece &&
                Capture(move) != Pawn &&
                !(context.extensions[ply-1] & RECAPTURE) &&
                  DestSquare(move) == DestSquare(context.last_move[ply-1])) {
                int capVal = PieceValue(Capture(context.last_move[ply-1]));
                int capVal2 = PieceValue(Capture(move));
                if( Util::Abs(capVal - capVal2) < PAWN_VALUE/2) {
                    extend += RECAPTURE_EXTENSION;
                    context.extensions[ply] |= RECAPTURE;
                    root->stats->recap_extensions++;
                }
            }
            if (options.search.threat_extensions && threat) {
                root->stats->threat_extensions++;
                extend += THREAT_EXTENSION;
            }
            if (limit) {
                extend /= 2;
            }
            if (extend > DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
            if (options.search.history_pruning && depth>=3*DEPTH_INCREMENT && 
               (node.hibound==node.alpha+1) && !in_check && extend==0 &&
                CanHPrune(move) && board.getMaterial(board.Side()).material_level()>3) {
               if (PieceMoved(move)!=Pawn || Rank(DestSquare(move),board.Side())<6) {
                  extend -= HISTORY_REDUCTION;
                  context.extensions[ply] |= HPRUNE;
               }
            }
#ifdef RAZORING
            if (razor && (node.hibound==node.alpha+1) && 
                TypeOfMove(move) != Promotion &&
                (Capture(move) == EmptyPiece) && extend == 0 &&
                in_check_after_move == NotInCheck) {
                extend += RAZOR_EXTENSION;
                root->stats->num_razored++;
#ifdef _TRACE
                if (master()) {
                    indent(ply); cout << "razoring, new depth = " << depth+RAZOR_EXTENSION << endl;
                }
#endif
            }
#endif
#ifdef FUTILITY_PRUNE
            int fmax;
            if (in_check_after_move == NotInCheck && 
		(node.hibound==node.alpha+1) && 
                !threat &&
                pruneOk(board,node.alpha,depth+extend,in_check,move,fmax,mat_score)) {
                est = PieceValue(Capture(move));
                if (fmax + est < node.alpha) {
#ifdef _TRACE
                    if (master()) {
                         indent(ply); cout << "fwd pruned." << endl;
                    }
#endif
                    ++fpruned_moves;
                    continue;
                }
            }
#endif
#ifdef SMP
            if (!first && root->iteration_depth >= THREAD_SPLIT_START &&
                ply <= THREAD_SPLIT_LIMIT) {
               // try to search this move in a different thread
               ThreadInfo *ti;
               if ((ti = controller->pool.checkOut(&node)) != NULL) {
                  // a thread is available, split off a new search,
                  // copying state so it can run independently:
#ifdef THREAD_TRACE
                  cout << "splitting" << endl;
#endif
                  Search *newSearch = new Search(board,this,&node,
                      move,extend);
                  // start searching in the new thread:
                  ti->start(newSearch);
#ifdef SMP_STATS
                  root->stats->splits++;
#endif
                  continue;
               }
            }
#endif
            board.DoMove(move);
            if (!in_check && board.any_attacks(board.KingPos(
            board.OppositeSide()),board.Side())) {
#ifdef _TRACE
                if (master()) {
                  indent(ply); cout << "Illegal move!" << endl;
                }
#endif
                board.UndoMove(move,context.states[ply]);
                continue;
            }
            // in rare cases we may not have applied the check extension
            // before the move, so do it here
            if (in_check_after_move == CheckUnknown) {
                in_check_after_move = board.CheckStatus();
                if (in_check_after_move == InCheck && extend == 0) {
                    context.extensions[ply] |= CHECK;
                    extend += CHECK_EXTENSION;
                    root->stats->check_extensions++;
                    if (limit) extend /= 2;
                    if (extend > DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
                }
            }
            ASSERT(in_check_after_move != CheckUnknown);
            ASSERT((in_check_after_move == InCheck) ==
                board.any_attacks(board.KingPos(
                board.Side()),board.OppositeSide()));
            board.set_check_status(in_check_after_move);
            if (depth+extend-DEPTH_INCREMENT >= 0) {
                try_score = -search(board, -node.hibound, -node.alpha,
                    ply+1, depth+extend-DEPTH_INCREMENT);
            }
            else {
                try_score = -quiesce(board, -node.hibound, -node.alpha,
                    ply+1, 0);
            }
            if (try_score != Illegal && !first && (try_score > node.alpha) && (try_score < node.beta)) {
                // We failed to get a cutoff and must re-search
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "score = " << try_score << " - no cutoff, researching .." << endl;
                   indent(ply); cout << "window = [" << node.alpha << "," << node.beta
                       << "]" << endl;
                }
#endif
                if (depth+extend-DEPTH_INCREMENT >= 0)
                    try_score=-search(board,
                        -node.beta,-node.alpha,ply+1,depth+extend-DEPTH_INCREMENT);
                else
                    try_score=-quiesce(board,-node.beta,-node.alpha,ply+1,0);
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << ply << ". ";
                    MoveImage(move,cout);
                    cout << ' ' << try_score << endl;
                }
#endif
            }
            // verify history pruning
            if ((context.extensions[ply] & HPRUNE) && try_score >= node.beta) {
              // verify history pruning
              if (depth+extend-DEPTH_INCREMENT >= 0) {
                  try_score = -search(board, -node.beta, -node.alpha,
                     ply+1, depth-DEPTH_INCREMENT);
               }
               else {
                   try_score = -quiesce(board, -node.beta, -node.alpha,
                      ply+1, 0);
               }
            }
            board.UndoMove(move,context.states[ply]);
            if (context.terminate) {
	      break;
            }
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << ' ' << try_score;
                if (first) cout << " (pv)";
                cout << endl;
            }
#endif
            if (try_score != Illegal) {
                first = 0;
                num_try++;
                if (try_score > node.alpha) {
                   Lock(node.nodeLock);
                   // don't use root version
                   Search::update_move(node,move,try_score);
                   Unlock(node.nodeLock);
                   if (try_score > node.beta) {
                     node.cutoff++;
		     break;
		   }
		}
                if (try_score >= Constants::BIG-1-ply) {
                   break;                        // mating move found
                }
                // zero-width window for the rest of the moves
                node.hibound = node.alpha+1;
            }
        }                                         // end move loop
#ifdef _TRACE
        if (node.score >= node.beta && master()) {
            indent(ply);
            cout << "**CUTOFF**" << endl;
        }
#endif
#ifdef SMP
    // We may have other threads still searching.
    if (node.cutoff || context.terminate) {
       // We just need to stop the other threads.
       // We don't care about their results.
       stopAllThreads(&node);
       waitForThreadCompletion(&node);
    }
    else {
       // The other threads have possibly useful info for us.
       // Wait until they complete:
       waitForThreadCompletion(&node);
    }
#endif
    if (context.terminate) return node.alpha;
    if (num_try == 0
#ifdef FUTILITY_PRUNE
            && fpruned_moves == 0
#endif
        ) {
            // no moves were tried
            if (in_check) {
#ifdef _TRACE
                    if (master()) {
                       indent(ply); cout << "mate" << endl;
                    }
#endif
                    node.score = -(Constants::BIG - ply);
                    // we will not have a move so do not adjust alpha
                    //if (node.score > node.alpha) node.alpha = node.score;
                    goto pv_update;
            }
            else {                                // stalemate
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "stalemate!" << endl;
                }
#endif
                node.score = draw_score(board);
                if (node.score > node.alpha) node.alpha = node.score;
                goto pv_update;                   // skip the hash insert
            }
        }
        else if (num_try > 0) {
           if (Capture(context.last_move[ply]) == EmptyPiece &&
               PromoteTo(context.last_move[ply]) == InvalidPiece) {
                if (node.score >=node.beta) Move_Generator::set_killer(context.last_move[ply], ply);
                Move_Generator::update_history(board,context.last_move[ply],node.score>=node.beta);
            }
#ifdef MOVE_ORDER_STATS
            if (master()) {
                root->stats->move_order_count++;
                if (best_count<4) {
                    root->stats->move_order[best_count]++;
                }
            }
#endif
        }
    }

    search_end:
    // don't insert into the hash table if we are terminating - we may
    // not have an accurate score.
    if (controller->use_hash_table && !context.terminate) {
        if (node.best == NullMove) node.best = hash_move;
        // store the position in the hash table, if there's room
        int value = node.score;
        Position_Info::ValueType val_type;
        // Adjust mate scores to reflect current ply. But only
        // if the score is in bounds.
        if (value > initial_alpha && value < node.beta) {
            if (value < -Constants::BIG+200) {
                value -= ply;
            }
            else if (value > Constants::BIG-200) {
                value += ply;
            }
        }
#ifdef _TRACE
        char typeChar;
#endif
        if (value <= initial_alpha) {
            val_type = Position_Info::UpperBound;
#ifdef _TRACE
            typeChar = 'U';
#endif
            // We don't have a "best" move, because all moves
            // caused alpha cutoff.  But if there was a hash
            // move or an initial move produced by internal
            // interative deepening, save it in the hash table
            // so it will be tried first next time around.
            node.best = hash_move;
        }
        else if (value >= node.beta) {
            val_type = Position_Info::LowerBound;
#ifdef _TRACE
            typeChar = 'L';
#endif
        }
        else {
#ifdef FUTILITY_PRUNE
            val_type = fpruned_moves ? Position_Info::Invalid : Position_Info::Valid;
#else
            val_type = Position_Info::Valid;
#endif
#ifdef _TRACE
            typeChar = 'E';
#endif
        }
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "storing type=" << typeChar <<
                " ply=" << ply << " depth=" << depth << " value=" << value <<
                " move=";
            MoveImage(node.best,cout); 
            cout << endl;
        }
#endif
        Hash_Entry new_pos(board.HashCode(rep_count), depth,
            controller->age,
            val_type, 0,
            board.CheckStatus() == InCheck,
            node.score, initial_alpha+1<node.beta,node.best);
        Lock(hash_lock);
        switch (storeHash(new_pos)) {
            case 0: hash_inserts++; break;
            case 1: hash_replaces++; break;
            case 2: hash_inserts_failed++; break;
            default:  break;
        }
        Unlock(hash_lock);
    }
    pv_update:
#ifdef FUTILITY_PRUNE
    root->stats->futility_pruning += fpruned_moves;
#endif
    if (!context.terminate && initial_alpha != node.alpha && node.score < node.beta) {
        if (num_try == 0 || IsNull(node.best))
           root->pv_length[ply] = 0;
        else
            update_pv(board,node.best,ply);
    }
    ASSERT(node.score >= -Constants::BIG && node.score <= Constants::BIG);
    return node.score;
}


// variant of search() used for IID
int Search::dsearch(Board & board, int alpha, int beta,
                   int ply, int depth) {
    if (master() && root->time_check_counter >= Time_Check_Interval) {
        root->time_check_counter = 0;
        context.terminate = root->check_time(board,ply,0,root->iteration_depth);
    }
    ASSERT(ply < Constants::MaxPly);
    root->time_check_counter++;
    num_nodes++;
    context.best[ply] = NullMove;                         // move we will return
    if (context.terminate) {
        return alpha;
    }
    int in_check = 0;
    int threat = 0;
    int rep_count;
#ifdef FUTILITY_PRUNE
    int est;
#endif
    // At this point we need to know if we are in check or not.
    in_check = board.CheckStatus(context.last_move[ply-1]) == InCheck;
#ifdef _DEBUG	
	if (!in_check) {
		if(board.calc_attacks(board.KingPos(board.Side()),board.OppositeSide())!=Bitmap(0)) {
			cout << board << endl;
			MoveImage(context.last_move[ply-1],cout);
			cout << endl;
			ASSERT(0);
		}
	}
#endif
#ifdef _TRACE
    if (master() && in_check) { indent(ply); cout << "in_check=" << in_check << endl;}
#endif
#if defined(FUTILITY_PRUNE) || defined(RAZORING)
    int mat_score=-Constants::BIG;
#endif
#if defined(FUTILITY_PRUNE)
    int fpruned_moves=0;
#endif
#ifdef RAZORING
    int razor = 0;
    if (!in_check && board.getMaterial(board.OppositeSide()).piece_count() >= 3
        &&  depth <= RAZOR_DEPTH && depth > RAZOR_DEPTH+RAZOR_EXTENSION) {
        if (mat_score == -Constants::BIG)
            mat_score = Scoring::material_score(board);
        razor = (mat_score + RAZOR_MARGIN) < alpha;
    }
#endif

	rep_count = board.rep_count();
#ifdef MOVE_ORDER_STATS
    int best_count = 0;
#endif
    int move_count = 0;
    int num_try = 0;                              // # of legal moves tried

    NodeInfo node;
    node.alpha = node.score = alpha; node.beta = node.hibound = beta;
    node.ply = ply; node.depth = depth;

    memcpy(&context.states[ply],&board.state,sizeof(Board_State));

    // Null move cutoff does not help us get a move, so skip it here.
    // Do IID here (recursively), if possible
    Move hash_move = NullMove;
    if (depth > 2*DEPTH_INCREMENT) {
        int d;
        d = Util::Min(depth-2*DEPTH_INCREMENT,depth/2);
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "== deepening, depth = " << d
                << endl;
        }
#endif
        // call special search routine. Return value not significant.
        dsearch(board,alpha,beta,ply,d);
        if (context.terminate) {
            return alpha;
        }
#ifdef _TRACE
        if (master()) { 
           indent(ply); cout << "== deepening done.";
        }
#endif
        hash_move = context.best[ply];
#ifdef _DEBUG
        if (!IsNull(hash_move)) {
          ASSERT(legal_move(board,hash_move));
        }
#endif
#ifdef _TRACE
        if (master()) {
            if (!IsNull(hash_move)) {
                indent(ply); cout << "  hash_move = ";
                MoveImage(hash_move,cout);
            }
            cout << endl;
        }
#endif
    }
    {
        Move_Generator mg(board, ply, NullMove, master());
        int try_score;
        // place to taper off extensions:
        const int limit = ply > limits[root->iteration_depth];
        //
        // Now we are ready to loop through the moves from this position
        //
        int first = 1;
        while (!node.cutoff) {
            Move move = in_check ? mg.NextEvasion() : mg.NextMove();
            if (IsNull(move)) break;
            if (IsUsed(move)) continue;
            move_count++;
            num_moves++;
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "trying " << ply << ". ";
                MoveImage(move,cout);
                if (first) cout << "(pv)";
                cout << " [" << node.alpha << "," << node.beta << "]";
                cout << endl;
            }
#endif
            context.last_move[ply] = move;
            CheckStatusType in_check_after_move = board.wouldCheck(move);
            int extend = 0;
            // see if we should apply any extensions at this node.
            context.extensions[ply] = 0;
            if (options.search.repetition_extensions && rep_count==1) {
                context.extensions[ply] |= REPETITION;
                extend += REPETITION;
                root->stats->rep_extensions++;
            }
            if (options.search.forced_extensions && mg.forced()) {
                // one single reply to check
                extend += FORCED_EXTENSION;
                root->stats->forced_extensions++;
            }
            else if (options.search.check_extensions && in_check_after_move == InCheck) {
                context.extensions[ply] |= CHECK;
                extend += CHECK_EXTENSION;
                root->stats->check_extensions++;
            }
            else if (options.search.pawn_push_extensions && 
                PieceMoved(move) == Pawn &&
                (Rank(DestSquare(move),board.Side())==7)) {
                    context.extensions[ply] |= PAWN_PUSH;
                    root->stats->pawn_extensions++;
            }
            else if (options.search.recapture_extensions &&
                Capture(move) != EmptyPiece &&
                Capture(move) != Pawn &&
                !(context.extensions[ply-1] & RECAPTURE) &&
                  DestSquare(move) == DestSquare(context.last_move[ply-1])) {
                int capVal = PieceValue(Capture(context.last_move[ply-1]));
                int capVal2 = PieceValue(Capture(move));
                if( Util::Abs(capVal - capVal2) < PAWN_VALUE/2) {
                    extend += RECAPTURE_EXTENSION;
                    context.extensions[ply] |= RECAPTURE;
                    root->stats->recap_extensions++;
                }
            }
            if (options.search.threat_extensions && threat) {
                root->stats->threat_extensions++;
                extend += THREAT_EXTENSION;
            }
            if (limit) {
                extend /= 2;
            }
            if (extend > DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
            if (options.search.history_pruning && depth>=3*DEPTH_INCREMENT && 
                (node.hibound==node.alpha+1) && !in_check && extend==0 &&
                CanHPrune(move) && board.getMaterial(board.Side()).material_level()>3) {
               if (PieceMoved(move)!=Pawn || Rank(DestSquare(move),board.Side())<6) {
                  extend -= HISTORY_REDUCTION;
                  context.extensions[ply] |= HPRUNE;
               }
            }
#ifdef RAZORING
            if (razor && (node.hibound==node.alpha+1) &&
                TypeOfMove(move) != Promotion &&
                (Capture(move) == EmptyPiece) && extend == 0 &&
                in_check_after_move == NotInCheck) {
                extend += RAZOR_EXTENSION;
                root->stats->num_razored++;
#ifdef _TRACE
                if (master()) {
                    indent(ply); cout << "razoring, new depth = " << depth+RAZOR_EXTENSION << endl;
                }
#endif
            }
#endif
#ifdef FUTILITY_PRUNE
            int fmax;
	    // do not futility prune at 1st move
            if (in_check_after_move == NotInCheck && 
		(node.hibound==node.alpha+1) &&
                !threat &&
                pruneOk(board,node.alpha,depth+extend,in_check,move,fmax,mat_score)) {
               est = PieceValue(Capture(move));
               if (fmax + est < node.alpha) {
#ifdef _TRACE
                   if (master()) {
                       indent(ply); cout << "fwd pruned." << endl;
                   }
#endif
                   ++fpruned_moves;
                   continue;
               }
            }
#endif
#ifdef SMP
            if (!first && root->iteration_depth >= THREAD_SPLIT_START &&
                ply <= THREAD_SPLIT_LIMIT) {
               // try to search this move in a different thread
               ThreadInfo *ti;
               if ((ti = controller->pool.checkOut(&node)) != NULL) {
                  // a thread is available, split off a new search,
                  // copying state so it can run independently:
#ifdef THREAD_TRACE
                  cout << "splitting" << endl;
#endif
                  Search *newSearch = new Search(board,this,&node,
                      move,extend);
                  // start searching in the new thread:
                  ti->start(newSearch);
#ifdef SMP_STATS
                  root->stats->splits++;
#endif
                  continue;
               }
            }
#endif
            board.DoMove(move);
            if (!in_check && board.any_attacks(board.KingPos(
            board.OppositeSide()),board.Side())) {
#ifdef _TRACE
                if (master()) {
                  indent(ply); cout << "Illegal move!" << endl;
                }
#endif
                board.UndoMove(move,context.states[ply]);
                continue;
            }
            // in rare cases we may not have applied the check extension
            // before the move, so do it here
            if (in_check_after_move == CheckUnknown) {
                in_check_after_move = board.CheckStatus();
                if (in_check_after_move == InCheck && extend == 0) {
                    context.extensions[ply] |= CHECK;
                    extend += CHECK_EXTENSION;
                    root->stats->check_extensions++;
                    if (limit) extend /= 2;
                    if (extend > DEPTH_INCREMENT) extend = DEPTH_INCREMENT;
                }
            }
            ASSERT(in_check_after_move != CheckUnknown);
            ASSERT((in_check_after_move == InCheck) ==
                board.any_attacks(board.KingPos(
                board.Side()),board.OppositeSide()));
            board.set_check_status(in_check_after_move);
            if (depth+extend-DEPTH_INCREMENT >= 0) {
                try_score = -search(board, -node.hibound, -node.alpha,
                    ply+1, depth+extend-DEPTH_INCREMENT);
            }
            else {
                try_score = -quiesce(board, -node.hibound, -node.alpha,
                    ply+1, 0);
            }
            if (try_score != Illegal && !first && (try_score > node.alpha) && (try_score < node.beta)) {
                // We failed to get a cutoff and must re-search
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "score = " << try_score << " - no cutoff, researching .." << endl;
                   indent(ply); cout << "window = [" << node.alpha << "," << node.beta
                       << "]" << endl;
                }
#endif
                if (depth+extend-DEPTH_INCREMENT >= 0)
                    try_score=-search(board,
                        -node.beta,-node.alpha,ply+1,depth+extend-DEPTH_INCREMENT);
                else
                    try_score=-quiesce(board,-node.beta,-node.alpha,ply+1,0);
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << ply << ". ";
                    MoveImage(move,cout);
                    cout << ' ' << try_score << endl;
                }
#endif
            }
            // verify history pruning
            if ((context.extensions[ply] & HPRUNE) && try_score >= node.beta) {
              // verify history pruning
              if (depth+extend-DEPTH_INCREMENT >= 0) {
                  try_score = -search(board, -node.beta, -node.alpha,
                     ply+1, depth-DEPTH_INCREMENT);
               }
               else {
                   try_score = -quiesce(board, -node.beta, -node.alpha,
                      ply+1, 0);
               }
            }
            board.UndoMove(move,context.states[ply]);
            if (context.terminate) {
	      break;
            }
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << ' ' << try_score;
                if (first) cout << " (pv)";
                cout << endl;
            }
#endif
            if (try_score != Illegal) {
                first = 0;
                num_try++;
                if (try_score > node.alpha) {
                   Lock(node.nodeLock);
                   // don't use root version
                   Search::update_move(node,move,try_score);
                   Unlock(node.nodeLock);
                   if (try_score > node.beta) {
                     node.cutoff++;
		     break;
		   }
                }
                if (try_score >= Constants::BIG-1-ply) {
                   break;                        // mating move found
                }
                // zero-width window for the rest of the moves
                node.hibound = node.alpha+1;
            }
        }                                         // end move loop
#ifdef _TRACE
        if (node.score >= node.beta && master()) {
            indent(ply);
            cout << "**CUTOFF**" << endl;
        }
#endif
#ifdef SMP
    // We may have other threads still searching.
    if (node.cutoff || context.terminate) {
       // We just need to stop the other threads.
       // We don't care about their results.
       stopAllThreads(&node);
       waitForThreadCompletion(&node);
    }
    else {
       // The other threads have possibly useful info for us.
       // Wait until they complete:
       waitForThreadCompletion(&node);
    }
#endif
    if (context.terminate) return node.alpha;
    if (num_try == 0
#ifdef FUTILITY_PRUNE
            && fpruned_moves == 0
#endif
        ) {
            // no moves were tried
            if (in_check) {
#ifdef _TRACE
                    if (master()) {
                       indent(ply); cout << "mate" << endl;
                    }
#endif
                    node.score = -(Constants::BIG - ply);
                    if (node.score > node.alpha) node.alpha = node.score;
                    goto pv_update;
            }
            else {                                // stalemate
#ifdef _TRACE
                if (master()) {
                   indent(ply); cout << "stalemate!" << endl;
                }
#endif
                node.score = draw_score(board);
                if (node.score > node.alpha) node.alpha = node.score;
                goto pv_update;                   // skip the hash insert
            }
        }
        else if (num_try > 0) {
           if (Capture(context.last_move[ply]) == EmptyPiece &&
               PromoteTo(context.last_move[ply]) == InvalidPiece) {
                if (node.score >=node.beta) Move_Generator::set_killer(context.last_move[ply], ply);
                Move_Generator::update_history(board,context.last_move[ply],node.score>=node.beta);
            }
#ifdef MOVE_ORDER_STATS
            if (master()) {
                root->stats->move_order_count++;
                if (best_count<4) {
                    root->stats->move_order[best_count]++;
                }
            }
#endif
        }
    }

    pv_update:
#ifdef FUTILITY_PRUNE
    root->stats->futility_pruning += fpruned_moves;
#endif
    // no pv update - screws up parent pv
    ASSERT(node.score >= -Constants::BIG && node.score <= Constants::BIG);
    return node.score;
}


