// Copyright 1994-1997, 2005 by Jon Dart.  All Rights Reserved.

#ifndef _SEARCH_H
#define _SEARCH_H

#include "board.h"
#include "movegen.h"
#include "hash.h"
#include "stats.h"
#ifdef SMP
#include "threadp.h"
#endif
#include <time.h>
#include <set>
using namespace std;

class RootSearch;
class SearchController;

// There are 4 levels of verbosity.  Silent mode does no output to
// the console - it is used by the Windows GUI version.  Debug
// level is used to output debug info.  Whisper is used to "whisper"
// comments on a chess server.  "Trace" is used in trace mode
// (-t) of the chess server client.
enum Talk_Level { Silent, Debug, Whisper, Trace };
    
enum Search_Type { Fixed_Depth, Time_Limit, Fixed_Time };

typedef void (*PostFunction)(const Statistics &);
typedef int (*TerminateFunction)(const Statistics &);

struct NodeInfo {
    NodeInfo() : cutoff(0),best(NullMove) {
      LockInit(nodeLock);
    }
    virtual ~NodeInfo () {
      LockDestroy(nodeLock);
    }
    int score;
    int alpha, beta, hibound, ply, depth;
    int cutoff;
    Move best;
#ifdef SMP
    LockDefine(nodeLock);
    // set of active threads associated with this node:
    set <ThreadInfo *> workThreads;
#endif
};

struct RootNodeInfo : public NodeInfo {
  int move_index, move_count;
  Move *moves;
};

struct SearchContext {

    SearchContext();
    virtual ~SearchContext();
#ifdef SMP
    ThreadInfo *ti;
#endif
    Board board;
    int terminate;
    Move best[Constants::MaxPly];
    Move last_move[Constants::MaxPly];
    int extensions[Constants::MaxPly];
    int extend; 
    Board_State states[Constants::MaxPly];
};

class Search {

    // Encapsulates alpha-beta search routine. 

    enum Extension_Type { RECAPTURE=1, CHECK=2, PAWN_PUSH=4, CAPTURE=8,
       FORCED=16, REPETITION=32, HPRUNE=64 };

public:
        
    Search(const Board &board, Search *parent);

    Search(const Board &board, Search *parent, SearchController *c);

    Search(const Board &board, Search *parent, NodeInfo *node,
      Move m,int extension);

    virtual ~Search();
    
    int is_draw( const Board &board, int &rep_count, int ply );
    
    int draw_score(const Board & board);
    
    int search(Board & board, int alpha, int beta,
               int ply, int depth);

    int dsearch(Board & board, int alpha, int beta,
               int ply, int depth);

    int master() const { return parent == NULL; }

#ifdef SMP
    // perform a subsidiary search in a separate thread
    void searchSMP(ThreadInfo *);
#endif

    int quiesce(Board & board, int alpha, int beta,
                int ply, int depth);

    int evalu8(const Board & board, int alpha, int beta, int ply);

    void update_pv(const Board &, const Move &m,int ply);

    Move getBest() const;

    Search * split(const Board &,int ply);

    void terminate() {
       context.terminate = 1;
    }

    NodeInfo *getParentNode() const {
      return parentNode;
    }
    
    virtual int update_move(NodeInfo &, Move move, int score);

    SearchContext context;

protected:
    Move moveToMake;
    Search *parent;
    NodeInfo *parentNode;
    SearchController *controller;
    RootSearch *root;

    void init(const Board &,Search *parent,SearchController *);
    void cleanup();
    
#ifdef SMP
    void stopAllThreads(NodeInfo *);
    void waitForThreadCompletion(NodeInfo *);
#endif
private:
    // make this class uncopyable:
    Search( const Search & );
    Search &operator = ( const Search & );

    int qsearch_check(Board & board, int alpha, int beta,
                int ply, int depth);

    int qsearch_no_check(Board & board, int alpha, int beta,
                int ply, int depth);

};

class RootSearch : public Search {
    friend class Search;
    friend class SearchController;
public: 
    RootSearch(const Board &board, SearchController *c);

    virtual ~RootSearch();

    Move ply0_search(const Board &board, 
                Search_Type srcType,
                int32 timelimit,
                int32 xtratime,
                int plylimit,
                const int background,
                Statistics &stat_buf,
                Talk_Level t,
                Move *exclude, int num_exclude);

    time_t getTimeAdded() {
        return time_added;
    }

    // Get the starting time for the search in centiseconds
    time_t get_start_time() const
    {
        return start_time;
    }
    
    Search_Type get_search_type() const
    {
        return type_of_search;
    }
    
    void set_time_limit(time_t limit,time_t xtra_time_limit)
    {
       type_of_search = Time_Limit;
       time_target = limit;
       xtra_time = xtra_time_limit;
    }
    
    void set_ply_limit(int ply)
    {
       type_of_search = Fixed_Depth;
       ply_limit = ply;
    }
    
    time_t get_time_limit() const
    { 
        return time_target + (time_added ? xtra_time : 0);
    }
    
    int is_background_search() const
    {
        return bkgrnd_search;
    }
    
    void set_background(int b);

    virtual int update_move(NodeInfo &rootNode, Move move, int score);

    void terminate_now();
    
    int was_terminated() const;

private:
    void init(const Board &);

    int check_time(const Board &,int ply, int rank, int interation_depth);
        
    // output something to the console
    void show_status(const Board &board,const Move bestMove);

    void iterate();

    int ply0_search(Board &board, int alpha, int beta,
            int depth, Move exclude [], int num_exclude, int multipv_count);

    void update_stats(int alpha, int beta,int multipv_count);

    Move excludes[MAX_PV];
    int num_excludes;
    Search_Type type_of_search;
    int easy;
    Move easy_move;
    Move ply0first; // first move to search
    Move ply0moves[Constants::MaxMoves];
    int ply0move_count;
    Move pv[Constants::MaxPly][Constants::MaxPly];
    int pv_length[Constants::MaxPly];
    ColorType computerSide;
    ColorType side_to_move;
    Statistics *stats;
    time_t time_target, start_time, xtra_time, time_added;
    int time_check_counter;
    int ply_limit;
    int bkgrnd_search;
    int root_total_mat;
    int null_depth;
    int terminated, time_up;
    Board initial_board;
    StateType state;
    int iteration_depth;
    int display_score; // The score displayed to the user
    LockDefine(pv_lock);
};

class SearchController {
    // Manages shared state that is not particular to a search
    // thread
    friend class Search;
    friend class RootSearch;

public:

    SearchController();

    virtual ~SearchController();

    Move find_best_move( 
            const Board &board, 
            Search_Type srcType,
            int32 time_limit,
            int32 xtratime,
            int ply_limit,
            int background,
            int uci,
            Statistics &stats,
            Talk_Level t);
    
    Move find_best_move( 
            const Board &board, 
            Search_Type srcType,
            int32 time_limit,
            int32 xtratime,
            int ply_limit,
            int background,
            int uci,
            Statistics &stats,
            Talk_Level t,
            Move *exclude,
            int num_excludes);
    
    void register_command_function(void (*cmd)(char *,int &)) {
       command_function = cmd;
    }
     
    void register_pending_function(void (*cmd)(const char *)) {
       pending_function = cmd;
    }
     
    PostFunction register_post_function(PostFunction post) {
       PostFunction tmp = post_function;
       post_function = post;
       return tmp;
    }
     
    TerminateFunction register_terminate_function(TerminateFunction term) {
       TerminateFunction tmp = terminate_function;
       terminate_function = term;
       return tmp;
    }
     
    void set_background(int b);
    
    void init();

    void reinit();

    void setRatingDiff(int rdiff);

    // check console input
    int check_input(const Board &);

    // return # of nodes visited in current search
    uint64 get_numnodes() const;
            
    Talk_Level get_talk_level() const {
       return talkLevel;
    }

    void set_talk_level(Talk_Level t)
    {
       talkLevel = t;
    }
    
    int draw_score(const Board & board, ColorType computerSide,
       int root_total_mat);
    
    void set_time_limit(time_t limit,time_t xtra_time_limit) {
       rootSearch->set_time_limit(limit,xtra_time_limit);
    }

    void terminate_now() const {
       if (rootSearch) rootSearch->terminate_now();
    }

    int was_terminated() const {
       return (rootSearch) ? rootSearch->terminated : 0;
    }

private:
    void cleanup(int clearHash);

    // pointer to function, called to process a command
    // received during the search (console programs only).
    void (*command_function)(char *,int &);

    // pointer to function, called to output status during
    // a search (used in console mode only).
    void (*post_function)(const Statistics &);

    int (*terminate_function)(const Statistics &);

    // Function to add a command to the deferred execution
    // stack (UNIX_ONLY)
    void (*pending_function)(const char*);

    // Calculate the recommended hash table size (expressed
    // in number of entries).  
    static long calc_hash_size();
    
    int age;
    int explicit_excludes;
    int use_hash_table;
    int hash_table_size;
    int ratingDiff;
    int ratingFactor; 
    int pipe;
    int terminated;
    int uci;
    Talk_Level talkLevel;

    RootSearch *rootSearch;

#ifdef SMP
    ThreadPool pool;
#endif
};

#endif
