// Copyright 2005 by Jon Dart. All Rights Reserved.

#include "threadp.h"
#include "search.h"
#include <iostream>
#ifndef _WIN32
#include <errno.h>
#include <fcntl.h>
#endif
using namespace std;

#ifdef _WIN32
static DWORD WINAPI parkingLot(void *x)
#else
static void * CDECL parkingLot(void *x)
#endif
{
   ThreadInfo *ti = (ThreadInfo*)x;
   while (ti->state != ThreadInfo::Terminating) {
#ifdef _THREAD_TRACE
      cout << "thread in parkingLot" << endl;
#endif
      ti->state = ThreadInfo::Idle; // mark thread available again
#ifdef _WIN32
      WaitForSingleObject(ti->hEvent1,INFINITE);
#else
      struct sembuf sb = { 0, 0, 0 };
      // wait for semaphore value 0  
      int err;
      if ((err = semop(ti->semid,&sb,1)) != 0) {
        if (err == EINTR || err == EIDRM) break;
        else perror("semop");
      }
#endif
#ifdef _THREAD_TRACE
      cout <<"thread unblocked" << endl;
#endif
      if (ti->state == ThreadInfo::Terminating) break;
#ifdef _THREAD_TRACE
      cout << "calling searchSMP" << endl;
#endif
      ASSERT(ti->work);
      ti->work->searchSMP(ti);
   }
   return 0;
}

ThreadInfo::~ThreadInfo() {
#ifdef _WIN32
  CloseHandle(hEvent1);
#else
  if (semctl(semid,0,IPC_RMID)) perror("semctl");
#endif
}

void ThreadInfo::start(Search *w) {
    ASSERT(w);
    work = w;
#ifdef _THREAD_TRACE
    cout << "start" << endl;
#endif
    // unblock the thread so it can execute
#ifdef _WIN32
    if (!SetEvent(hEvent1)) {
        cerr << "error: SetEvent" << endl;
    }
#else
    struct sembuf sb = { 0, -1, IPC_NOWAIT };
    // decrement semaphore value:
    if (semop(semid,&sb,1)) perror("semop");
#endif
}

ThreadInfo::ThreadInfo() 
 : state(Idle),
#ifdef _WIN32
   thread_id(NULL),
   work(NULL)
#else
   work(NULL),
   count(0)
#endif
{
#ifdef _THREAD_TRACE
  cout << "starting thread.." << endl;
#endif
#ifdef _WIN32
      // initially the controller blocks the thread
      hEvent1 = CreateEvent(NULL,FALSE,FALSE,NULL);
      DWORD id;
      thread_id = CreateThread(NULL,0,
          parkingLot,this,
          0,
          &id);
#else
      semid = semget(IPC_PRIVATE,1,IPC_CREAT | 0777);
      if (semid == -1) {
	perror("semaphore creation failed");
        return;
      }
#ifdef _THREAD_TRACE
      cout << "semid = " << semid << endl;
#endif
      union semun val;
      // initialize semaphore to 1
      val.val = 1; 
      if (semctl(semid, 0, SETVAL, val)) {
        perror("semctl");
        return;
      }
      if (pthread_create(&thread_id, NULL, parkingLot, this)) {
	perror("thread creation failed");
      }
#endif
}

ThreadPool::ThreadPool(int n) 
: nthreads(n) {
   data = new ThreadInfo *[n];
   LockInit(poolLock);
   for (int i = 0; i < n; i++) {
      data[i] = new ThreadInfo();
   }
}

ThreadPool::~ThreadPool() {
    shutDown();
}

void ThreadPool::shutDown() {
    for (int i = 0; i < nthreads; i++) {
       if (data[i]->state == ThreadInfo::Idle) {
          data[i]->state = ThreadInfo::Terminating;
#ifdef _WIN32
          SetEvent(data[i]->hEvent1);
#else
       struct sembuf sb = { 0, -1, IPC_NOWAIT };
       // decrement semaphore value:
       if (semop(data[i]->semid,&sb,1)) perror("semop");
#endif
       }
       else {
#ifdef _WIN32
          WaitForSingleObject(data[i]->thread_id,INFINITE);
#else
          void *value_ptr;
          pthread_join(data[i]->thread_id,&value_ptr);
#endif
       }
       delete data[i];
    }
    delete [] data; 
}

ThreadInfo * ThreadPool::checkOut(NodeInfo *forNode) {
    Lock(poolLock);
    Lock(forNode->nodeLock);
    for (int i = 0; i < nthreads; i++) {
#ifdef _WIN32
       if (data[i]->state == ThreadInfo::Idle) {
#else
	 // verify the thread is available and is actually
	 // at the synchronization point
       if (data[i]->state == ThreadInfo::Idle &&
           semctl(data[i]->semid,0,GETZCNT)>0) {
#endif
           data[i]->state = ThreadInfo::Working;
           // keep track of active threads for the calling node:
           data[i]->work = NULL;
           data[i]->parentNode = forNode;
           forNode->workThreads.insert(data[i]);
           ASSERT(forNode->workThreads.size()==1);
           Unlock(forNode->nodeLock);
           Unlock(poolLock);
#ifdef _THREAD_TRACE
           cout <<"checkOut" << endl;
#endif
           return data[i]; 
       }
    }
    // no luck, no free threads
    Unlock(forNode->nodeLock);
    Unlock(poolLock);
    return NULL;
}

void ThreadPool::checkIn(ThreadInfo *ti) {
#ifdef _THREAD_TRACE
    cout << "checkIn" << endl;
#endif
    Lock(poolLock);
    Lock(ti->parentNode->nodeLock);
    ASSERT(ti->parentNode->workThreads.size());
    // dissociate the thread from the parent
    ti->parentNode->workThreads.erase(ti);
    delete ti->work;
    Unlock(ti->parentNode->nodeLock);
    Unlock(poolLock);
#ifdef _WIN32
    if (!ResetEvent(ti->hEvent1))
        cerr << "ResetEvent failed" << endl;
#else
    // increment semaphore value so we will wait
    // when entering loop again
    struct sembuf sb = { 0, 1, IPC_NOWAIT };
    if (semop(ti->semid,&sb,1)) perror("semop");
#endif
}

int ThreadPool::activeCount() {
    Lock(poolLock);
    int count = 0;
    for (int i = 0; i < nthreads; i++) {
       if (data[i]->state == ThreadInfo::Working) ++count;
    }
    Unlock(poolLock);
    return count;
}


